'''
'''

from .sim_base import Simulator

import numpy as np
import traits.api as tr

from .hist import Hist
from .i_hist import IHist
from .i_tloop import ITLoop
from .i_tstep import ITStep
import bmcs_utils.api as bu

@tr.provides(ITStep)
class TStep(bu.Model):
    '''Manage the data and metadata of a time step within an interation loop.
    '''
    title = tr.Str('<unnamed>')

    tloop_type = tr.Type(ITLoop)
    '''Type of time loop to be used with the model
    '''

    #=========================================================================
    # HISTORY
    #=========================================================================
    hist_type = tr.Type(Hist)

    hist = tr.Property(bu.Instance(IHist))
    r'''History representation of the model response.
    '''
    @tr.cached_property
    def _get_hist(self):
        return self.hist_type(tstep_source=self)

    debug = tr.Bool(False)

    t_n1 = tr.Float(0.0, auto_set=False, enter_set=True)
    '''Target value of the control variable.
    '''
    U_n = tr.Float(0.0, auto_set=False, enter_set=True)
    '''Current fundamental value of the primary variable.
    '''
    U_k = tr.Float(0.0, auto_set=False, enter_set=True)
    '''Current trial value of the primary variable.
    '''

    def init_state(self):
        '''Initialize state.
        '''
        self.U_n = 0.0
        self.t_n1 = 0.0
        self.U_k = 0.0

    def record_state(self):
        '''Provide the current state for history recording.
        '''
        pass

    _corr_pred = tr.Property(depends_on='U_k,t_n1')

    @tr.cached_property
    def _get__corr_pred(self):
        return self.get_corr_pred(self.U_k, self.t_n1)

    R = tr.Property

    def _get_R(self):
        R, _ = self._corr_pred
        return R

    dR = tr.Property

    def _get_dR(self):
        _, dR = self._corr_pred
        return dR

    R_norm = tr.Property

    def _get_R_norm(self):
        R = self.R
        return np.sqrt(R * R)

    def make_iter(self):
        d_U = self.R / self.dR
        self.U_k += d_U

    def make_incr(self):
        '''Update the control, primary and state variables..
        '''
        self.U_n = self.U_k
        # self.hist.record_timestep()

    sim = tr.Property()
    '''Launch a simulator - currently only one simulator is allowed
    for a model. Mutiple might also make sense when different solvers
    are to be compared. The simulator pulls the time loop type
    from the model.
    '''
    @tr.cached_property
    def _get_sim(self):
        return Simulator(self)



'''

@author: rch
'''


from traits.api import \
    HasStrictTraits, Instance, Button, Event, \
    DelegatesTo
from traits.etsconfig.api import ETSConfig

from .bmcs_study import BMCSStudy

if ETSConfig.toolkit == 'wx':
    from traitsui.wx.tree_editor import \
        DeleteAction
if ETSConfig.toolkit == 'qt4':
    from traitsui.qt4.tree_editor import \
        DeleteAction
else:
    raise ImportError("tree actions for %s toolkit not available" %
                      ETSConfig.toolkit)


# tree_node = TreeNode(node_for=[BMCSRootNode, BMCSTreeNode],
#                      auto_open=False,
#                      children='tree_node_list',
#                      label='node_name',
#                      view='tree_view',
#                      menu=Menu(plot_self, DeleteAction),
#                      )
#
# leaf_node = TreeNode(node_for=[BMCSLeafNode],
#                      auto_open=True,
#                      children='',
#                      label='node_name',
#                      view='tree_view',
#                      menu=Menu(plot_self)
#                      )
#
# tree_editor = TreeEditor(
#     nodes=[tree_node, leaf_node],
#     selected='selected_node',
#     orientation='vertical'
# )


class BMCSWindow(BMCSStudy):

    selected_node = Instance(HasStrictTraits)

    def _selected_node_changed(self):
        self.selected_node.ui = self

    def get_vot_range(self):
        return self.viz_sheet.get_vot_range()

    vot = DelegatesTo('viz_sheet')

    data_changed = Event

    replot = Button

    def _replot_fired(self):
        self.figure.clear()
        self.selected_node.plot(self.figure)
        self.data_changed = True

    clear = Button()

    def _clear_fired(self):
        self.figure.clear()
        self.data_changed = True



from .i_bcond import \
    IBCond
from traits.api import List,  Any, provides
from bmcs_utils.api import Float, Int, Enum, Instance, View, Item
from ibvpy.view.ui import BMCSTreeNode
from ibvpy.tfunction import TimeFunction, TFMonotonic

import numpy as np


@provides(IBCond)
class BCDof(BMCSTreeNode):
    '''
    Implements the IBC functionality for a constrained dof.

    Example of a complex constraint:

    For example the specification

    BCDof( var = 'f',
           value = 0.,
           dof = 2,
           link_dofs = [3,4],
           link_coeffs = [0.5,0.5] )

    means that

    U[2] = 0.5*U[3] + 0.5*U[4] Note that U[2] is non-zero

    and is regarded as a natural boundary condition.

    On the the other hand, the specification

    cos(alpha) * U[2] + sin(slpha) * U[3] = 0.4

    can be expressed as for U[2] as
    U[2] = - sin(alpha) / cos(alpha) * U[3] + 0.4 / cos(alpha)
    so that the corresponding BCDof specification has the form

    BCDof( var = 'u',
           value = 0.4 / cos(alpha),
           dof = 2,
           link_dofs = [3],
           link_coeffs = [-sin(alpha)/cos(alpha) ] )

    '''
    node_name = 'boundary condition'

    name = 'BC DOF'
    var = Enum(options=['u', 'f', 'eps', 'sig'],
               label='Variable',
               BC=True
               )
    dof = Int(label='Degree of freedom',
              BC=True,
              )
    value = Float(label='Value',
                  BC=True,
                  )
    link_dofs = List(Int,
                     BC=True,
                     label='Linear dependencies',
                     tooltip='Degrees of freedom linked\n'
                     'with the current by link coefficients')
    '''
    List of dofs that determine the value of the current dof
    
    If this list is empty, then the current dof is
    prescribed. Otherwise, the dof value is given by the
    linear combination of DOFs in the list (see the example below)
    '''
    link_coeffs = List(Float,
                       BC=True,
                       label='Link coefficients',
                       tooltip='Multipliers for linear combination\n'
                       'equation')
    '''
    Coefficients of the linear combination of DOFs specified in the
    above list.
    '''

    time_function = Instance(TimeFunction)
    def _time_function_default(self):
        return TFMonotonic()

    '''
    Time function prescribing the evolution of the boundary condition.
    '''

    ipw_view = View(
        Item('var'),
        Item('dof'),
        Item('value'),
        Item('time_function')
    )

    def is_essential(self):
        return self.var == 'u'

    def is_linked(self):
        return self.link_dofs != []

    def is_constrained(self):
        '''
        Return true if a DOF is either explicitly prescribed or it depends on other DOFS.
        '''
        return self.is_essential() or self.is_linked()

    def is_natural(self):
        return self.var == 'f' or self.var == 'eps' or self.var == 'sig'

    def get_dofs(self):
        return [self.dof]

    def setup(self, sctx):
        '''
        Locate the spatial context.
        '''
        return

    _constraint = Any

    def register(self, K):
        '''Register the boundary conditions in the equation system.
        '''
        if self.is_essential():
            a = self.dof   # affected dof
            alpha = np.array(self.link_coeffs, np.float64)
            n_ix_arr = np.array(list(self.link_dofs), dtype=int)
            self._constraint = K.register_constraint(a=a, u_a=self.value,
                                                     alpha=alpha, ix_a=n_ix_arr)

    def apply(self, step_flag, sctx, K, R, t_n, t_n1):
        '''
        According to the kind specification add the
        '''
        if self.is_essential():

            # The displacement is applied only in the first iteration step!.
            #
            if step_flag == 'predictor':
                try:
                    ua_n = self.value * float(self.time_function(t_n))
                    ua_n1 = self.value * float(self.time_function(t_n1))
                except ValueError as ex:
                    print('t_n', t_n, t_n1)
                    raise ex
                u_a = ua_n1 - ua_n
            elif step_flag == 'corrector':
                u_a = 0

            self._constraint.u_a = u_a

        elif self.is_natural():

            R_a = self.value * float(self.time_function(t_n1))

            # Add the value to the proper equation.
            #
            # if a is involved in another essential constraint, redistribute
            # it according to the link coefficients infolved in that constraint!
            #
            R[self.dof] += R_a

            if self.is_linked():

                # Prepare the indexes and index arrays
                #
                n = self.link_dofs  # constraining dofs
                n_ix = np.ix_(n)   # constraining dofs as array

                # Distribute the load contribution to the proportionally loaded dofs
                #
                alpha = np.array(self.link_coeffs, np.float64)
                R[n_ix] += alpha.transpose() * R_a

    def update_plot(self, axes):
        self.time_function.update_plot(axes)

if __name__ == '__main__':

    from ibvpy.mesh.fe_grid import FEGrid
    from ibvpy.mesh.fe_domain import FEDomain
    from ibvpy.fets.fets1D.fets1D2l import FETS1D2L
    from ibvpy.api import \
        TStepper as TS, RTDofGraph, TLoop, \
        TLine
    from ibvpy.tmodel.mats1D.mats1D_elastic.mats1D_elastic import MATS1DElastic

    fets_eval = FETS1D2L(mats_eval=MATS1DElastic(E=10., A=1.))

    # Discretization
    fe_domain1 = FEGrid(coord_max=(10., 0., 0.),
                        shape=(10,),
                        fets_eval=fets_eval)

    fe_domain2 = FEGrid(coord_min=(10., 0., 0.),
                        coord_max=(20., 0., 0.),
                        shape=(10,),
                        fets_eval=fets_eval)

    fe_domain = FEDomain(subdomains=[fe_domain1, fe_domain2])
    ts = TS(dof_resultants=True,
            sdomain=fe_domain,
            bcond_list=[BCDof(var='u', dof=0, value=0.),
                        BCDof(
                            var='u', dof=5, link_dofs=[16], link_coeffs=[1.], value=0.),
                        BCDof(var='f', dof=21, value=10)],
            rtrace_list=[RTDofGraph(name='Fi,right over u_right (iteration)',
                                    var_y='F_int', idx_y=0,
                                    var_x='U_k', idx_x=1),
                         ]
            )

    # Add the time-loop control
    tloop = TLoop(tstepper=ts, tline=TLine(min=0.0, step=1, max=1.0))

    ts.set(sdomain=FEDomain(subdomains=[fe_domain1, fe_domain2]))

    ts.set(bcond_list=[BCDof(var='u', dof=0, value=0.),
                       BCDof(
                           var='u', dof=5, link_dofs=[16], link_coeffs=[1.], value=0.),
                       BCDof(var='f', dof=21, value=10)])
    print(tloop.eval())




from numpy import array, sign, argwhere

from ibvpy.api import FEDomain, FERefinementGrid, FEGrid, TStepBC as TS, \
    RTraceDomainListField, BCDof, BCSlice
from ibvpy.api import TLoop, TLine
from ibvpy.mesh.fe_ls_domain import FELSDomain
from ibvpy.tmodel.mats1D.mats1D_elastic.mats1D_elastic import MATS1DElastic
from ibvpy.fets.fets1D.fets1D2l import FETS1D2L
from ibvpy.fets.fets1D.fets1D2l3u import FETS1D2L3U
from ibvpy.fets.fets_ls.fets_crack import FETSCrack
from numpy import ma

if __name__ == '__main__':
    def example_1d():
        fets_eval = FETS1D2L3U( mats_eval = MATS1DElastic( E = 20. ) )
        xfets_eval = FETSCrack( parent_fets = fets_eval,
                                int_order = 2 )

        # Discretization

        fe_domain = FEDomain()
        fe_level1 = FERefinementGrid( domain = fe_domain, fets_eval = fets_eval )
        fe_grid1 = FEGrid( coord_max = ( 2., 0., 0. ),
                           shape = ( 2, ),
                           fets_eval = fets_eval,
                           level = fe_level1 )


        enr = True
        if enr:
            fe_xdomain = XFESubDomain( domain = fe_domain,
                                       fets_eval = xfets_eval,
                                       #fe_grid_idx_slice = fe_grid1[1,0],
                                       fe_grid_slice = fe_grid1['X  - .75'] )
            fe_xdomain.deactivate_sliced_elems()

        ts = TS( dof_resultants = True,
                 sdomain = fe_domain,
                 bcond_list = [BCSlice( var = 'u', value = -1. / 2., dims = [0],
                                        slice = fe_grid1[ 0, 0 ] ),
                                BCSlice( var = 'u', value = 0., dims = [0],
                                        slice = fe_grid1[ -1, -1 ] ),
                                        ],
                 rtrace_list = [
#                                 RTDofGraph(name = 'Fi,right over u_right (iteration)' ,
#                                       var_y = 'F_int', idx_y = 0,
#                                       var_x = 'U_k', idx_x = 1),
                            RTraceDomainListField( name = 'Stress' ,
                                 var = 'eps', idx = 0, warp = True ),
                             RTraceDomainListField( name = 'Displacement' ,
                                            var = 'u', idx = 0,
                                            warp = True ),
#                                     RTraceDomainField(name = 'N0' ,
#                                                  var = 'N_mtx', idx = 0,
#                                                  record_on = 'update')
                        ]
                    )
#        
#        # Add the time-loop control
        tloop = TLoop( tstepper = ts,
                       debug = True,
                       tolerance = 1e-4, RESETMAX = 0,
                       tline = TLine( min = 0.0, step = 1, max = 1.0 ) )

        #print "elements ",fe_xdomain.elements[0]
        if enr:
            print('parent elems ', fe_xdomain.fe_grid_slice.elems)
            print('parent dofs ', fe_xdomain.fe_grid_slice.dofs)
            print("dofmap ", fe_xdomain.elem_dof_map)
            print("ls_values ", fe_xdomain.dots.dof_node_ls_values)
            print('intersection points ', fe_xdomain.fe_grid_slice.r_i)#
            print("triangles ", fe_xdomain.dots.int_division)
            print('ip_coords', fe_xdomain.dots.ip_coords)
            print('ip_weigths', fe_xdomain.dots.ip_weights)
            print('ip_offset ', fe_xdomain.dots.ip_offset)
            print('ip_X_coords', fe_xdomain.dots.ip_X)
            print('ip_ls', fe_xdomain.dots.ip_ls_values)
            print('vtk_X ', fe_xdomain.dots.vtk_X)
            print('vtk triangles ', fe_xdomain.dots.rt_triangles)
            print("vtk data ", fe_xdomain.dots.get_vtk_cell_data( 'blabla', 0, 0 ))
            print('vtk_ls', fe_xdomain.dots.vtk_ls_values)
            print('J_det ', fe_xdomain.dots.J_det_grid)

        tloop.eval()

        from ibvpy.plugins.ibvpy_app import IBVPyApp
        ibvpy_app = IBVPyApp( ibv_resource = ts )
        ibvpy_app.main()

    def example_2d():
        from ibvpy.api import FEDomain, FERefinementGrid, FEGrid, TStepper as TS, \
            BCDofGroup, RTraceDomainListField
        from ibvpy.api import TLoop, TLine
        from ibvpy.mesh.xfe_subdomain import XFESubDomain
        from ibvpy.tmodel.mats2D.mats2D_elastic.mats2D_elastic import MATS2DElastic
        from ibvpy.tmodel.mats2D import MATS2DPlastic
        from ibvpy.fets.fets2D.fets2D4q import FETS2D4Q
        from ibvpy.fets.fets2D import FETS2D9Q
        from ibvpy.fets.fets2D.fets2D4q8u import FETS2D4Q8U
        from ibvpy.fets.fets_ls.fets_crack import FETSCrack
        #fets_eval = FETS2D4Q( mats_eval = MATS2DPlastic( E = 1., nu = 0. ) )
        fets_eval = FETS2D4Q8U( mats_eval = MATS2DPlastic( E = 1., nu = 0. ) )
        xfets_eval = FETSCrack( parent_fets = fets_eval,
                                int_order = 5,
                                tri_subdivision = 1 )

        # Discretization

        fe_domain = FEDomain()
        fe_level1 = FERefinementGrid( domain = fe_domain, fets_eval = fets_eval )
        fe_grid1 = FEGrid( coord_max = ( 1., 1. ),
                           shape = ( 8, 8 ),
                           fets_eval = fets_eval,
                           level = fe_level1 )

        #ls_function = lambda X, Y: X - Y - 0.13
        ls_function = lambda X, Y: ( X - 0.52 ) ** 2 + ( Y - 0.72 ) ** 2 - 0.51 ** 2
        bls_function = lambda X, Y:-( ( X - 0.5 ) ** 2 + ( Y - 0.21 ) ** 2 - 0.28 ** 2 )
        bls_function2 = lambda X, Y:-( ( X - 0.5 ) ** 2 + ( Y - 0.21 ) ** 2 - 0.38 ** 2 )

        # design deficits:
        # - How to define a level set spanned over several fe_grids
        #   (i.e. it is defined over the hierarchy of FESubDomains)
        # - Patching of subdomains within the FEPatchedGrid (FERefinementGrid)
        # - What are the compatibility conditions?
        # - What is the difference between FEGridLeveSetSlice
        #   and FELSDomain? 
        #   FELSDomain is associated with a DOTS - Slice is not.

        # FEGrid has a multidimensional array - elem_grid
        # it can be accessed through this index.
        # it is masked by the activity map. The activity map can 
        # be defined using slices and level sets.
        # the elems array enumerates the elements using the activity map.
        # in this way, the specialization of grids is available implicitly.
        #
        fe_xdomain = FELSDomain( domain = fe_domain,
                                 fets_eval = xfets_eval,
                                 fe_grid = fe_grid1,
                                 ls_function = ls_function,
                                 bls_function = bls_function,
                                 )

        fe_tip_xdomain = FELSDomain( domain = fe_domain,
                                     fets_eval = xfets_eval,
                                     fe_grid = fe_xdomain,
                                     ls_function = bls_function,
                                     )

        # deactivation must be done only after the dof enumeration has been completed
        fe_xdomain.deactivate_intg_elems_in_parent()
        fe_tip_xdomain.deactivate_intg_elems_in_parent()

        fe_xdomain.bls_function = bls_function2
        fe_tip_xdomain.ls_function = bls_function2

        # deactivation must be done only after the dof enumeration has been completed
        fe_xdomain.deactivate_intg_elems_in_parent()
        fe_tip_xdomain.deactivate_intg_elems_in_parent()

        #
        # General procedure:
        # 1) define the level sets with the boundaries
        # 2) use the bls to identify the tips of the level set
        # 3) use independent level sets to introduce indpendently junctions.
        #
        # get the extended dofs of the bls_elems and constrain it
        #
        cdofs = fe_tip_xdomain.elem_xdof_map.flatten()
        bc_list = [ BCDof( var = 'u', dof = dof, value = 0.0 )
                    for                   dof in cdofs ]

        # construct the time stepper

        ts = TS( dof_resultants = True,
                 sdomain = fe_domain,
                 bcond_list = [BCSlice( var = 'u', value = -0.1, dims = [1],
                                        slice = fe_grid1[:, 0, :, 0] ),
                                BCSlice( var = 'u', value = 0., dims = [0],
                                          slice = fe_grid1[:, 0, :, 0] ),
                                BCSlice( var = 'u', value = 0., dims = [0, 1],
                                         slice = fe_grid1[:, -1, :, -1] )
                                ] + bc_list,
                 rtrace_list = [
#                                 RTDofGraph(name = 'Fi,right over u_right (iteration)' ,
#                                       var_y = 'F_int', idx_y = 0,
#                                       var_x = 'U_k', idx_x = 1),
                            RTraceDomainListField( name = 'Stress' ,
                                 var = 'sig_app', idx = 0, warp = True ),
                             RTraceDomainListField( name = 'Displacement' ,
                                            var = 'u', idx = 0,
                                            warp = True ),
#                                     RTraceDomainField(name = 'N0' ,
#                                                  var = 'N_mtx', idx = 0,
#                                                  record_on = 'update')
                        ]
                    )
#        

        do = 'print'

        if do == 'print':

            p = 'state'

            if p == 'grids':
                print('fe_xdomain.ls mask')
                print(fe_xdomain.ls_mask)
                print('fe_xdomain.idx mask')
                print(fe_xdomain.idx_mask)
                print('fe_xdomain.intg mask')
                print(fe_xdomain.intg_mask)
                print('fe_xdomain.xelems_mask')
                print(fe_xdomain.xelems_mask)
                print('fe_xdomain.xelems_grid_ix')
                print(fe_xdomain.xelems_grid_ix)
                print('fe_xdomain.ls_elem_grid')
                print(fe_xdomain.ls_elem_grid)
                print('fe_xdomain.ls_ielem_grid')
                print(fe_xdomain.ls_ielem_grid)
                print('fe_xdomain.intg_elem_grid')
                print(fe_xdomain.intg_elem_grid)

                print('fe_tip_xdomain.ls_mask`')
                print(fe_tip_xdomain.ls_mask)
                print('fe_tip_xdomain.intg_mask`')
                print(fe_tip_xdomain.intg_mask)
                print('fe_tip_xdomain.idx_mask`')
                print(fe_tip_xdomain.idx_mask)
                print('fe_tip_xdomain.xelems_mask')
                print(fe_tip_xdomain.xelems_mask)
                print('fe_tip_xdomain.xelems_grid_ix')
                print(fe_tip_xdomain.xelems_grid_ix)
                print('fe_tip_xdomain.ls_elem_grid')
                print(fe_tip_xdomain.ls_elem_grid)
                print('fe_tip_xdomain.ls_ielems_grid')
                print(fe_tip_xdomain.ls_ielem_grid)
                print('fe_tip_xdomain.intg_elem_grid')
                print(fe_tip_xdomain.intg_elem_grid)

            if p == 'maps':

                print('fe_xdomain.elem_dof_map')
                print(fe_xdomain.elem_dof_map)
                print('fe_tip_xdomain.elem_dof_map')
                print(fe_tip_xdomain.elem_dof_map)

                print('fe_xdomain.elems')
                print(fe_xdomain.elems)
                print('fe_tip_xdomain.elems')
                print(fe_tip_xdomain.elems)

                print('fe_xdomain.elem_X_map')
                print(fe_xdomain.elem_X_map)
                print('fe_tip_xdomain.elem_X_map')
                print(fe_tip_xdomain.elem_X_map)

            if p == 'fields':

                print("ls_values ", fe_xdomain.dots.dof_node_ls_values)
                print("tip ls_values ", fe_tip_xdomain.dots.dof_node_ls_values)

                print('intersection points ', fe_xdomain.ls_intersection_r)
                print('tip intersection points ', fe_tip_xdomain.ls_intersection_r)

                print("triangles ", fe_xdomain.dots.rt_triangles)
                print("vtk points ", fe_xdomain.dots.vtk_X)
                print("vtk data ", fe_xdomain.dots.get_vtk_cell_data( 'blabla', 0, 0 ))

                print('ip_triangles', fe_xdomain.dots.int_division)
                print('ip_coords', fe_xdomain.dots.ip_coords)
                print('ip_weigths', fe_xdomain.dots.ip_weights)
                print('ip_offset', fe_xdomain.dots.ip_offset)
                print('ip_X_coords', fe_xdomain.dots.ip_X)
                print('ip_ls', fe_xdomain.dots.ip_ls_values)
                print('vtk_ls', fe_xdomain.dots.vtk_ls_values)
                print('J_det ', fe_xdomain.dots.J_det_grid)

            if p == 'state':

                # Add the time-loop control
                print('STATE: initial')

                print('fe_xdomain.old_dots.state_elem grid')
                print(fe_xdomain.dots.state_start_elem_grid)
                print('fe_tip_xdomain.old_dots.state_elem grid')
                print(fe_tip_xdomain.dots.state_start_elem_grid)
                print('fe_xdomain.old_dots.state_end_elem grid')
                print(fe_xdomain.dots.state_end_elem_grid)
                print('fe_tip_xdomain.old_dots.state_end_elem grid')
                print(fe_tip_xdomain.dots.state_end_elem_grid)

                fe_xdomain.dots.state_array[:] = 25.5
                print('state_array 25', fe_xdomain.dots.state_array)
                fe_tip_xdomain.dots.state_array[:] = 58

                bls_function3 = lambda X, Y:-( ( X - 0.5 ) ** 2 + ( Y - 0.21 ) ** 2 - 0.58 ** 2 )

                fe_xdomain.bls_function = bls_function3
                fe_tip_xdomain.ls_function = bls_function3

                print('STATE: changed')

                print('fe_xdomain.old_dots.state_elem grid')
                print(fe_xdomain.dots.state_start_elem_grid)
                print('fe_tip_xdomain.old_dots.state_elem grid')
                print(fe_tip_xdomain.dots.state_start_elem_grid)
                print('fe_xdomain.old_dots.state_end_elem grid')
                print(fe_xdomain.dots.state_end_elem_grid)
                print('fe_tip_xdomain.old_dots.state_end_elem grid')
                print(fe_tip_xdomain.dots.state_end_elem_grid)

                print('state_array 25', fe_xdomain.dots.state_array.shape)
                print('state_array 25', fe_xdomain.dots.state_array[570:])
                print('state_array 58', fe_tip_xdomain.dots.state_array.shape)


        elif do == 'ui':

            tloop = TLoop( tstepper = ts,
                           debug = False,
                           tolerance = 1e-4, KMAX = 3, RESETMAX = 0,
                           tline = TLine( min = 0.0, step = 1, max = 1.0 ) )

            tloop.eval()
            from ibvpy.plugins.ibvpy_app import IBVPyApp
            ibvpy_app = IBVPyApp( ibv_resource = ts )
            ibvpy_app.main()
    #            
    #    #    print ts.F_int
    #    #    print ts.rtrace_list[0].trace.ydata

    example_2d()



'''
Created on 12.01.2016

@author: Yingxiong
'''

from ibvpy.fets import FETSEval, IFETSEval

import numpy as np
import sympy as sp
from traits.api import Int, Array, \
    Property, cached_property, Float, List, provides, Type
import traits.api as tr


r_ = sp.symbols('r')

#=================================================
# 8 nodes isoparametric volume element (3D)
#=================================================
# generate shape functions with sympy
xi_1 = sp.symbols('xi_1')
xi_2 = sp.symbols('xi_2')
xi_3 = sp.symbols('xi_3')


#=========================================================================
# Finite element specification
#=========================================================================

N_xi_i = sp.Matrix([0.5 - xi_1 / 2., 0.5 + xi_1 / 2.], dtype=np.float64)

dN_xi_ir = sp.Matrix([[-1. / 2], [1. / 2]], dtype=np.float64)


@provides(IFETSEval)
class FETS1D52ULRH(FETSEval):

    '''
    Fe Bar 2 nodes, deformation
    '''

    debug_on = True

    A_m = Float(1, desc='matrix area [mm2]')
    A_f = Float(1, desc='reinforcement area [mm2]')
    P_b = Float(1,
                desc='perimeter of the bond interface [mm]')

    # Dimensional mapping
    dim_slice = slice(0, 1)

    n_nodal_dofs = Int(2)

    dof_r = Array(value=[[-1], [1]])
    geo_r = Array(value=[[-1], [1]])
    vtk_r = Array(value=[[-1.], [1.]])
    vtk_cells = [[0, 1]]
    vtk_cell_types = 'Line'

    dots_class = Type

    n_dof_r = Property
    '''Number of node positions associated with degrees of freedom. 
    '''
    @cached_property
    def _get_n_dof_r(self):
        return len(self.dof_r)

    n_e_dofs = Property
    '''Number of element degrees
    '''
    @cached_property
    def _get_n_e_dofs(self):
        return self.n_nodal_dofs * self.n_dof_r

    def _get_ip_coords(self):
        offset = 1e-6
        return np.array([[-1 + offset, 0., 0.], [1 - offset, 0., 0.]])

    def _get_ip_weights(self):
        return np.array([1., 1.], dtype=float)

    # Integration parameters
    #
    ngp_r = 2

    def get_N_geo_mtx(self, r_pnt):
        '''
        Return geometric shape functions
        @param r_pnt:
        '''
        r = r_pnt[0]
        N_mtx = np.array([[0.5 - r / 2., 0.5 + r / 2.]])
        return N_mtx

    def get_dNr_geo_mtx(self, r_pnt):
        '''
        Return the matrix of shape function derivatives.
        Used for the conrcution of the Jacobi matrix.
        '''
        return np.array([[-1. / 2, 1. / 2]])

    def get_N_mtx(self, r_pnt):
        '''
        Return shape functions
        @param r_pnt:local coordinates
        '''
        return self.get_N_geo_mtx(r_pnt)

    def get_dNr_mtx(self, r_pnt):
        '''
        Return the derivatives of the shape functions
        '''
        return self.get_dNr_geo_mtx(r_pnt)

    xi_m = Array(value=[[-1], [1]], dtype=np.float64)
    r_m = Array(value=[[-1], [1]], dtype=np.float64)
    w_m = Array(value=[1, 1], dtype=np.float64)

    Nr_i_geo = List([(1 - r_) / 2.0,
                     (1 + r_) / 2.0, ])

    dNr_i_geo = List([- 1.0 / 2.0,
                      1.0 / 2.0, ])

    Nr_i = Nr_i_geo
    dNr_i = dNr_i_geo

    N_mi_geo = Property()

    @cached_property
    def _get_N_mi_geo(self):
        return self.get_N_mi(sp.Matrix(self.Nr_i_geo, dtype=np.float64))

    dN_mid_geo = Property()

    @cached_property
    def _get_dN_mid_geo(self):
        return self.get_dN_mid(sp.Matrix(self.dNr_i_geo, dtype=np.float64))

    N_mi = Property(depends_on='Nr_i_geo')

    @cached_property
    def _get_N_mi(self):
        return self.get_N_mi(sp.Matrix(self.Nr_i, dtype=np.float64))

    dN_mid = Property(depends_on='dNr_i_geo')

    @cached_property
    def _get_dN_mid(self):
        return self.get_dN_mid(sp.Matrix(self.dNr_i, dtype=np.float64))

    def get_N_mi(self, Nr_i):
        return np.array([Nr_i.subs(r_, float(r))
                         for r in self.r_m], dtype=np.float64)[...,0]

    def get_dN_mid(self, dNr_i):
        dN_mdi = np.array([[dNr_i.subs(r_, float(r))]
                           for r in self.r_m], dtype=np.float64)
        return np.einsum('mdi->mid', dN_mdi)

    n_m = Property(depends_on='w_m')

    @cached_property
    def _get_n_m(self):
        return len(self.w_m)

    A_C = Property(depends_on='A_m,A_f')

    @cached_property
    def _get_A_C(self):
        return np.array((self.A_f, self.P_b, self.A_m), dtype=np.float64)

    def get_B_EmisC(self, J_inv_Emdd):
        fets_eval = self

        n_dof_r = fets_eval.n_dof_r
        n_nodal_dofs = fets_eval.n_nodal_dofs

        n_m = fets_eval.n_gp
        n_E = J_inv_Emdd.shape[0]
        n_s = 3
        #[ d, i]
        r_ip = fets_eval.ip_coords[:, :-2].T
        # [ d, n ]
        geo_r = fets_eval.geo_r.T
        # [ d, n, i ]
        dNr_geo = geo_r[:, :, None] * np.array([1, 1]) * 0.5
        # [ i, n, d ]
        dNr_geo = np.einsum('dni->ind', dNr_geo)

        # shape function for the unknowns
        # [ d, n, i]
        Nr = 0.5 * (1. + geo_r[:, :, None] * r_ip[None, :])
        dNr = 0.5 * geo_r[:, :, None] * np.array([1, 1])

        # [ i, n, d ]
        Nr = np.einsum('dni->ind', Nr)
        dNr = np.einsum('dni->ind', dNr)
        Nx = Nr
        # [ n_e, n_ip, n_dof_r, n_dim_dof ]
        dNx = np.einsum('eidf,inf->eind', J_inv_Emdd, dNr)

        B = np.zeros((n_E, n_m, n_dof_r, n_s, n_nodal_dofs), dtype='f')
        B_N_n_rows, B_N_n_cols, N_idx = [1, 1], [0, 1], [0, 0]
        B_dN_n_rows, B_dN_n_cols, dN_idx = [0, 2], [0, 1], [0, 0]
        B_factors = np.array([-1, 1], dtype='float64')
        B[:, :, :, B_N_n_rows, B_N_n_cols] = (B_factors[None, None, :] *
                                              Nx[:, :, N_idx])
        B[:, :, :, B_dN_n_rows, B_dN_n_cols] = dNx[:, :, :, dN_idx]

        return B

    def _get_B_EimsC(self, dN_Eimd, sN_Cim):

        n_E, _, _, _ = dN_Eimd.shape
        n_C, n_i, n_m = sN_Cim.shape
        n_s = 3
        B_EimsC = np.zeros(
            (n_E, n_i, n_m, n_s, n_C), dtype='f')
        B_EimsC[..., [1, 1], [0, 1]] = sN_Cim[[0, 1], :, :]
        B_EimsC[..., [0, 2], [0, 1]] = dN_Eimd[:, [0, 1], :, :]

        return B_EimsC

    shape_function_values = tr.Property(tr.Tuple)
    '''The values of the shape functions and their derivatives at the IPs
    '''
    @tr.cached_property
    def _get_shape_function_values(self):
        N_mi = np.array([N_xi_i.subs(list(zip([xi_1, xi_2, xi_3], xi)))
                         for xi in self.xi_m], dtype=np.float64)
        N_im = np.einsum('mi->im', N_mi[...,0])
        dN_mir_arr = [np.array(dN_xi_ir.subs(list(zip([xi_1], xi)))).astype(np.float64)
                      for xi in self.xi_m]
        dN_mir = np.array(dN_mir_arr, dtype=np.float64)
        dN_nir_arr = [np.array(dN_xi_ir.subs(list(zip([xi_1], xi)))).astype(np.float64)
                      for xi in self.vtk_r]
        dN_nir = np.array(dN_nir_arr, dtype=np.float64)
        dN_imr = np.einsum('mir->imr', dN_mir)
        dN_inr = np.einsum('nir->inr', dN_nir)
        return (N_im, dN_imr, dN_inr)

    N_im = tr.Property()
    '''Shape function values in integration poindots.
    '''

    def _get_N_im(self):
        return self.shape_function_values[0]

    dN_imr = tr.Property()
    '''Shape function derivatives in integration poindots.
    '''

    def _get_dN_imr(self):
        return self.shape_function_values[1]

    dN_inr = tr.Property()
    '''Shape function derivatives in visualization poindots.
    '''

    def _get_dN_inr(self):
        return self.shape_function_values[2]

    I_sym_abcd = tr.Array(np.float64)

    def _I_sym_abcd_default(self):
        delta = np.identity(3)
        return 0.5 * \
            (np.einsum('ac,bd->abcd', delta, delta) +
             np.einsum('ad,bc->abcd', delta, delta))


if __name__ == '__main__':
    fe = FETS1D52ULRH()
    print('dN_imr', fe.dN_imr)



from math import pi as Pi
from numpy import \
    array,  frompyfunc
from traits.api import \
    Array, Bool, Callable,  Float, HasTraits, \
    Instance,   Range,  on_trait_change,  \
    Dict, Property, cached_property,   \
    WeakRef, String, List, Constant, Str, Type, TraitError

from ibvpy.mathkit.mfn.mfn_polar.mfn_polar import MFnPolar
from .matsXD_cmdm_phi_fn import \
    IPhiFn, PhiFnStrainSoftening, PhiFnStrainHardening, PhiFnStrainHardeningLinear, \
    PhiFnGeneral, PhiFnGeneralExtended, PhiFnGeneralExtendedExp, PhiFnStrainHardeningBezier
from ibvpy.util.traits.either_type import EitherType


# Traits UI imports
# @todo change to mfn_polar
#-------------------------------------------------------------------------
#                                     VariedParam
#-------------------------------------------------------------------------
class VariedParam(HasTraits):

    """
    Association between the spatial function and material parameter.
    """
    mats_eval = WeakRef
    phi_fn = WeakRef(IPhiFn)

    varname = String

    reference_value = Property(Float)

    def _get_reference_value(self):
        return getattr(self.phi_fn, self.varname)

    switched_on = Bool(False)

    polar_fn = Instance(MFnPolar)

    @on_trait_change('switched_on')
    def reset_polar_fn(self):
        if self.switched_on:
            if self.mats_eval.mfn_class == None:
                raise TraitError('No class for function representation specified')
            self.polar_fn = self.mats_eval.mfn_class()
        else:
            self.polar_fn = None

    # vectorized form of polar_fn to returning an array of coefficients
    polar_fn_vectorized = Property(Callable, depends_on='polar_fn')

    @cached_property
    def _get_polar_fn_vectorized(self):
        if self.switched_on:
            return frompyfunc(self.polar_fn, 1, 1)
        else:
            return frompyfunc(lambda angle: 1., 1, 1)


class PolarDiscr(HasTraits):

    '''
    Manager of the microplane arrays.

    This class is responsible for the generation and initialization
    and state management of an array of microplanes. Additionally, it
    can perform the setup of damage function parameters using the
    value of the microplane integrator object.
    '''

    mfn_class = Type(None)
    #-------------------------------------------------------------------------
    # Common parameters for for isotropic and anisotropic damage function specifications
    #-------------------------------------------------------------------------
    n_mp = Range(0, 50, 6,
                 label='Number of microplanes',
                 auto_set=False)

    E = Float(34e+3,
              label="E",
              desc="Young's Modulus",

              auto_set=False, enter_set=True)
    nu = Float(0.2,
               label='nu',
               desc="Poison's ratio",
               auto_set=False, enter_set=True)

    c_T = Float(0.0,
                label='c_T',
                desc='fraction of tangential stress accounted on each microplane',
                auto_set=False, enter_set=True)

    #-------------------------------------------------------------------------
    # list of angles
    #-------------------------------------------------------------------------
    alpha_list = Property(Array, depends_on='n_mp')

    @cached_property
    def _get_alpha_list(self):
        return array([Pi / self.n_mp * (i - 0.5) for i in range(1, self.n_mp + 1)])

    #-------------------------------------------------------------------------
    # Damage function specification
    #-------------------------------------------------------------------------

    phi_fn = EitherType(klasses=[PhiFnGeneral,
                                 PhiFnGeneralExtended,
                                 PhiFnGeneralExtendedExp,
                                 PhiFnStrainSoftening,
                                 PhiFnStrainHardening,
                                 PhiFnStrainHardeningLinear,
                                 PhiFnStrainHardeningBezier])

    def _phi_fn_default(self):
        print('setting phi_fn default')
        return PhiFnStrainSoftening(polar_discr=self)

    def _phi_fn_changed(self):
        print('setting phi_fn changed')
        self.phi_fn.polar_discr = self

    varied_params = List(Str, [])

    #-------------------------------------------------------------------------
    # Management of spatially varying parameters depending on the value of mats_eval
    #-------------------------------------------------------------------------
    varpars = Dict

    def _varpars_default(self):
        return self._get_varpars()

    @on_trait_change('phi_fn,varied_params')
    def _update_varpars(self):
        self.varpars = self._get_varpars()

    def _get_varpars(self):
        '''
        reset the varpar list according to the current phi_fn object.
        '''
        params = self.phi_fn.identify_parameters()
        varset = {}
        for key in params:
            varset[key] = VariedParam(phi_fn=self.phi_fn,
                                      mats_eval=self,
                                      varname=key)
            if key in self.varied_params:
                varset[key].switched_on = True
        return varset

    varpar_list = Property(List(VariedParam), depends_on='varpars')

    @cached_property
    def _get_varpar_list(self):
        return [self.varpars[key]
                for key in self.phi_fn.identify_parameters()]

    # variable selectable in the table of varied params (just for viewing)
    current_varpar = Instance(VariedParam)

    def _current_varpar_default(self):
        if len(self.varpar_list) > 0:
            return self.varpar_list[0]
        return None

    @on_trait_change('phi_fn')
    def set_current_varpar(self):
        if len(self.varpar_list) > 0:
            self.current_varpar = self.varpar_list[0]

    #-------------------------------------------------------------------------
    # Get the damage state for all microplanes
    #-------------------------------------------------------------------------
    def get_phi_arr(self, sctx, e_max_arr):
        '''
        Return the damage coefficients
        '''
        # gather the coefficients for parameters depending on the orientation
        carr_list = [self.varpars[key].polar_fn_vectorized(self.alpha_list)
                     for key in self.phi_fn.identify_parameters()]
        # vectorize the damage function evaluation
        n_arr = 1 + len(carr_list)
        phi_fn_vectorized = frompyfunc(self.phi_fn.get_value, n_arr, 1)
        # damage parameter for each microplane
        return phi_fn_vectorized(e_max_arr, *carr_list)

    def get_polar_fn_fracture_energy_arr(self, sctx, e_max_arr):
        '''
        Return the fracture energy contributions
        '''
        carr_list = [self.varpars[key].polar_fn_vectorized(self.alpha_list)
                     for key in self.phi_fn.identify_parameters()]
        # vectorize the damage function evaluation
        n_arr = 1 + len(carr_list)
        integ_phi_fn_vectorized = frompyfunc(self.phi_fn.get_integ, n_arr, 1)
        return self.E * integ_phi_fn_vectorized(e_max_arr, *carr_list)





from math import sqrt as scalar_sqrt
from ibvpy.tmodel.mats2D.mats2D_eval import MATS2DEval
from numpy import \
    array, zeros, dot, \
    sqrt, vdot, \
    float64, diag
from scipy.linalg import inv, norm
from traits.api import \
    Array, Enum, \
    Int, Trait, \
    Dict, Property, cached_property
from bmcs_utils.api import \
    Item, View, Float
from ibvpy.util.traits.either_type import EitherType
from .yield_face2D import IYieldFace2D, J2, DruckerPrager, Gurson, CamClay

#---------------------------------------------------------------------------
# Material time-step-evaluator for Scalar-Damage-Model
#---------------------------------------------------------------------------
class MATS2DPlastic(MATS2DEval):
    '''
    Elastic Model.
    '''

    # implements(IMATSEval)

    #-------------------------------------------------------------------------
    # Parameters of the numerical algorithm (integration)
    #-------------------------------------------------------------------------

    stress_state = Enum("plane strain", "plane stress",)
    algorithm = Enum("closest point", "cutting plane")

    #-------------------------------------------------------------------------
    # Material parameters
    #-------------------------------------------------------------------------
    yf = EitherType(klasses=[J2, DruckerPrager, Gurson, CamClay],
                    label="Yield Face",
                    desc="Yield Face Definition"
                    )
    E = Float(210.0e+3,
              label="E",
              desc="Young's Modulus")
    nu = Float(0.2,
               label='nu',
               desc="Poison's ratio")
    K_bar = Float(0.,
                  label='K',
                  desc="isotropic softening parameter")
    H_bar = Float(0.,
                  label='H',
                  desc="kinematic softening parameter")
    tolerance = Float(1.0e-4,
                      label='TOL',
                      desc="tolerance of return mapping")

    max_iter = Int(20,
                   label='Iterations',
                   desc="maximal number of iterations")

    D_el = Property(Array(float), depends_on='E, nu')

    @cached_property
    def _get_D_el(self):
        if self.stress_state == "plane_stress":
            return self._get_D_plane_stress()
        else:
            return self._get_D_plane_strain()

    H_mtx = Property(Array(float), depends_on='K_bar, H_bar')

    @cached_property
    def _get_H_mtx(self):
        H_mtx = diag([self.K_bar, self.H_bar, self.H_bar, self.H_bar])
        return H_mtx

    # This event can be used by the clients to trigger an action upon
    # the completed reconfiguration of the material model
    #

    #-------------------------------------------------------------------------
    # View specification
    #-------------------------------------------------------------------------

    view_traits = View(
        Item('yf'),
        Item('E'),
        Item('nu'),
        Item('K_bar'),
        Item('H_bar'),
        Item('tolerance'),
        Item('max_iter'),
        Item('algorithm'),
    )

    #-------------------------------------------------------------------------
    # Evaluation - get the corrector and predictor
    #-------------------------------------------------------------------------

    def get_corr_pred(self, sctx, eps_app_eng, d_eps, tn, tn1):
        '''
        Corrector predictor computation.
        @param eps_app_eng input variable - engineering strain
        '''
        delta_gamma = 0.
        if sctx.update_state_on:
            # print "in us"
            eps_n = eps_app_eng - d_eps
            sigma, f_trial, epsilon_p, q_1, q_2 = self._get_state_variables(
                sctx, eps_n)

            sctx.mats_state_array[:3] = epsilon_p
            sctx.mats_state_array[3] = q_1
            sctx.mats_state_array[4:] = q_2

        diff1s = zeros([3])
        sigma, f_trial, epsilon_p, q_1, q_2 = self._get_state_variables(
            sctx, eps_app_eng)
        # Note: the state variables are not needed here, just gamma
        diff2ss = self.yf.get_diff2ss(eps_app_eng, self.E, self.nu, sctx)
        Xi_mtx = inv(inv(self.D_el) + delta_gamma * diff2ss * f_trial)
        N_mtx_denom = sqrt(dot(dot(diff1s, Xi_mtx), diff1s))
        if N_mtx_denom == 0.:
            N_mtx = zeros(3)
        else:
            N_mtx = dot(Xi_mtx, self.diff1s) / N_mtx_denom
        D_mtx = Xi_mtx - vdot(N_mtx, N_mtx)

        # print "sigma ",sigma
        # print "D_mtx ",D_mtx
        return sigma, D_mtx


    def _get_D_plane_stress(self):
        E = self.E
        nu = self.nu
        D_stress = zeros([3, 3])
        D_stress[0][0] = E / (1.0 - nu * nu)
        D_stress[0][1] = E / (1.0 - nu * nu) * nu
        D_stress[1][0] = E / (1.0 - nu * nu) * nu
        D_stress[1][1] = E / (1.0 - nu * nu)
        D_stress[2][2] = E / (1.0 - nu * nu) * (1.0 / 2.0 - nu / 2.0)
        return D_stress

    def _get_D_plane_strain(self):
        E = self.E
        nu = self.nu
        D_strain = zeros([3, 3])
        D_strain[0][0] = E * (1.0 - nu) / (1.0 + nu) / (1.0 - 2.0 * nu)
        D_strain[0][1] = E / (1.0 + nu) / (1.0 - 2.0 * nu) * nu
        D_strain[1][0] = E / (1.0 + nu) / (1.0 - 2.0 * nu) * nu
        D_strain[1][1] = E * (1.0 - nu) / (1.0 + nu) / (1.0 - 2.0 * nu)
        D_strain[2][2] = E * (1.0 - nu) / (1.0 + nu) / (2.0 - 2.0 * nu)
        return D_strain

    #-------------------------------------------------------------------------
    # Response trace evaluators
    #-------------------------------------------------------------------------

    def get_sig_norm(self, sctx, eps_app_eng):
        sig_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        return array([scalar_sqrt(sig_eng[0] ** 2 + sig_eng[1] ** 2)])

    def get_eps_p(self, sctx, eps_app_eng):
        # print "eps tracer ", sctx.mats_state_array[:3]
        return sctx.mats_state_array[:3]

    # Declare and fill-in the rte_dict - it is used by the clients to
    # assemble all the available time-steppers.
    #
    rte_dict = Trait(Dict)

    def _rte_dict_default(self):
        return {'sig_app': self.get_sig_app,
                'eps_app': self.get_eps_app,
                'sig_norm': self.get_sig_norm,
                'eps_p': self.get_eps_p}



from ibvpy.tmodel.mats_eval import \
    IMATSEval, MATSEval
import numpy as np
from traits.api import provides
import bmcs_utils.api as bu


@provides(IMATSEval)
class MATS3DIfcCumSlip(MATSEval):
    node_name = "Cumulative bond-slip law"

    state_var_shapes = dict(
        omega_N=(),
        r_N=(),
        omega_T=(),
        z_T=(),
        alpha_T_a=(2,),
        s_p_a=(2,)
    )

    def init(self, *args):
        r'''
        Initialize the state variables.
        '''
        for a in args:
            a[...] = 0

    algorithmic = bu.Bool(False)

    def get_corr_pred(self, u_b, t_n1, omega_N, r_N, omega_T,
                      z_T, alpha_T_a, s_p_a):
        w = u_b[..., 0]
        s_a = u_b[..., 1:]
        # For normal - distinguish tension and compression
        T = w > 0.0
        H_w_N = np.array(w <= 0.0, dtype=np.float64)
        E_alg_N = H_w_N * self.E_N
        sig_N = self.E_N * w
        delta_lambda = 0.5 * self.E_N * (w * T) ** 2.0 * self.Ad * (1 + r_N) ** 2
        omega_N += delta_lambda
        r_N -= delta_lambda
        sig_N[T] = (1.0 - omega_N[T]) * self.E_N * (w[T])

        E_T = self.E_T

        tau_a = E_T * (s_a - s_p_a)
        Z = self.K_T * z_T
        X_a = self.gamma_T * alpha_T_a
        q_a = tau_a - X_a
        norm_q = np.sqrt(np.einsum('...na,...na->...n', q_a, q_a))
        f_trial = norm_q - self.tau_pi_bar - Z + self.a * sig_N

        I = np.where(f_trial > 1e-6)

        omega_T_I = omega_T[I]
        q_a_I = q_a[I]
        norm_q_I = norm_q[I]
        delta_lambda_I = f_trial[I] / \
                         (E_T / (1.0 - omega_T_I) + self.gamma_T + self.K_T)

        s_p_a[I, :] += (
                               delta_lambda_I / (1.0 - omega_T_I) / norm_q_I
                       )[:, np.newaxis] * q_a_I

        s_el_a_I = s_a[I] - s_p_a[I]
        Y_I = 0.5 * E_T * np.einsum('...a,...a->...', s_el_a_I, s_el_a_I)

        omega_T[I] += (
                (1.0 - omega_T[I]) ** self.c_T *
                delta_lambda_I * (Y_I / self.S_T) ** self.r_T
        )
        # * \
        # (self.tau_pi_bar / (self.tau_pi_bar - self.a * sig_N)
        alpha_T_a[I, :] += (
                (delta_lambda_I / norm_q_I)[:, np.newaxis] * q_a_I
        )
        z_T += delta_lambda

        tau_a[I, :] = (
                ((1 - omega_T[I]) * self.E_T)[:, np.newaxis] * s_el_a_I
        )
        # Secant stiffness
        E_alg_T = ((1 - omega_T) *
                   self.E_T)[:, np.newaxis, np.newaxis]
        # Consistent tangent operator
        sig = np.zeros_like(u_b)
        sig[..., 0] = sig_N
        sig[..., 1:] = tau_a
        N_ab = np.zeros((3, 3), dtype=np.float64)
        N_ab[0, 0] = 1
        T_ab = np.zeros((3, 3), dtype=np.float64)
        T_ab[(1, 2), (1, 2)] = 1
        E_NT = (np.einsum('ab,...->...ab', N_ab, E_alg_N) +
                np.einsum('ab,...ab->...ab', T_ab, E_alg_T))

        tau_a = E_T * (s_a - s_p_a)
        Z = self.K_T * z_T
        X_a = self.gamma_T * alpha_T_a
        q_a = tau_a - X_a
        norm_q = np.sqrt(np.einsum('...na,...na->...n', q_a, q_a))
        f_trial = norm_q - self.tau_pi_bar - Z + self.a * sig_N
        print('f_trial', f_trial[np.where(f_trial > 1e-6)])

        print('sig', sig)
        return sig, E_NT

    def _get_var_dict(self):
        var_dict = super(MATS3DIfcCumSlip, self)._get_var_dict()
        var_dict.update(
            slip=self.get_slip,
            s_el=self.get_s_el,
            shear=self.get_shear,
            omega=self.get_omega,
            s_pi=self.get_s_pi,
            alpha=self.get_alpha,
            z=self.get_z
        )
        return var_dict

    def get_slip(self, u_r, tn1, **state):
        return self.get_eps(u_r, tn1)[..., 0]

    def get_shear(self, u_r, tn1, **state):
        return self.get_sig(u_r, tn1, **state)[..., 0]

    def get_omega(self, u_r, tn1, s_pi, alpha, z, omega):
        return omega

    def get_s_pi(self, u_r, tn1, s_pi, alpha, z, omega):
        return s_pi

    def get_alpha(self, u_r, tn1, s_pi, alpha, z, omega):
        return alpha

    def get_z(self, u_r, tn1, s_pi, alpha, z, omega):
        return z

    def get_s_el(self, u_r, tn1, **state):
        s = self.get_slip(u_r, tn1, **state)
        s_p = self.get_s_pi(u_r, tn1, **state)
        s_e = s - s_p
        return s_e

    # ---------------------------------------
    # Tangential constitutive law parameters
    # ---------------------------------------
    E_T = bu.Float(100, label='E_T',
                   desc='Shear modulus of the interface',
                   MAT=True,
                   enter_set=True, auto_set=False)

    gamma_T = bu.Float(5000.,
                       label="Gamma_T",
                       desc=" Tangential Kinematic hardening modulus",
                       enter_set=True,
                       auto_set=False)

    K_T = bu.Float(10.0,
                   label="K_T",
                   desc="Tangential Isotropic harening",
                   enter_set=True,
                   auto_set=False)

    S_T = bu.Float(0.0001,
                   label="S_T",
                   desc="Damage strength",
                   enter_set=True,
                   auto_set=False)

    r_T = bu.Float(1.2,
                   label="r_T",
                   desc="Damage cumulation parameter",
                   enter_set=True,
                   auto_set=False)

    c_T = bu.Float(1.0,
                   label="c_T",
                   desc="Damage cumulation parameter",
                   enter_set=True,
                   auto_set=False)

    tau_pi_bar = bu.Float(2.0,
                          label="Tau_bar_T",
                          desc="Reversibility limit",
                          enter_set=True,
                          auto_set=False)

    a = bu.Float(0.0,
                 label="a",
                 desc="Lateral pressure coefficient",
                 enter_set=True,
                 auto_set=False)

    # -------------------------------------------
    # Normal_Tension constitutive law parameters (without cumulative normal strain)
    # -------------------------------------------
    E_N = bu.Float(100, label='E_N',
                   desc='Normal stiffness of the interface',
                   MAT=True,
                   enter_set=True, auto_set=False)

    Ad = bu.Float(10000.0,
                  label="Ad",
                  desc="Brittleness coefficient",
                  enter_set=True,
                  auto_set=False)

    traits_view = bu.View(
        bu.Item('E_T'),
        bu.Item('E_N'),
        bu.Item('gamma_T'),
        bu.Item('K_T'),
        bu.Item('S_T'),
        bu.Item('r_T'),
        bu.Item('c_T'),
        bu.Item('tau_pi_bar'),
        bu.Item('a'),
        bu.Item('Ad'),
    )



'''

   1 - recursivity
   2 - figures
   3 - sort
   
   package dependency - window depends on reporter
   parametric studies?
'''
from .reporter import Reporter
from traits.api import \
    Float,\
    Instance, Int

from .report_item import \
    RInputRecord, RInputSection, ROutputSection


class CrossSection(RInputSection):
    name = 'Cross section'
    A_m = Float(0.4, CS=True, unit='mm',
                label='Area of the matrix',
                math_symbol='A_\mathrm{m}')

    A_f = Float(0.4, CS=True, unit='mm',
                label='Area of the reinforcement',
                math_symbol='A_\mathrm{f}')

    P_b = Float(0.2, CS=True, unit='mm',
                label='Contact perimeter',
                math_symbol='P_\mathrm{b}')


class Geometry(RInputSection):
    name = 'Geometry'
    length = Float(10.0, GEO=True, unit='mm', symbol='$L$')
    width = Float(0.4, GEO=True, unit='mm', symbol='$b$')


class ModelInput(RInputSection):
    name = 'part1'
    title = 'Model part 1'
    desc = 'Testing p-study'
    n_e = Int(20, GEO=True)
    cross_section = Instance(CrossSection, (), report=True)
    geo = Instance(Geometry, (), report=True)


class ModelOutput(ROutputSection):
    model_input = Instance(ModelInput)
    name = 'part1'
    title = 'Model part 1'
    desc = 'Testing p-study'

    def get_records(self):
        return []


def generate_report():
    mpi = ModelInput(name='Part 1')

    mpo = ModelOutput()
    r = Reporter()
    r.report_items = [mpi, mpo
                      ]
    rinput_tree = r.yield_rinput_traits()

    for path, trait in list(rinput_tree.items()):
        print(path, trait)
        trait.value_range = [2, 3, 4, 5]

    r.write()
    r.show_tex()
    r.run_pdflatex()
    r.show_pdf()


if __name__ == '__main__':
    rinput_tree = generate_report()




#-- Imports --------------------------------------------------------------------

from os.path \
    import join, dirname
    
from numpy \
    import sqrt
    
from numpy.random \
    import random

from traits.api \
    import HasTraits, Property, Array, Any, Event, \
    on_trait_change, Instance, WeakRef, Int, Str, Bool, Trait

from ibvpy.view.plot3d.mayavi_util.pipelines import \
    MVPolyData, MVPointLabels, MVStructuredGrid    

class ElemView(HasTraits):
    '''Get the element numbers.
    '''
    elem_num = Int(-1)
    elem_coords = Array

#-- ElemArrayView Class Definition -------------------------------------------------

class ElemArrayView ( HasTraits ):

    data = Array
    rt_domain = WeakRef
    
    
    mvp_elem_labels = Trait( MVPointLabels )
    def _mvp_elem_labels_default(self):
        return MVPointLabels( name = 'Geo node numbers', 
                                  points = self._get_current_elem_coords,
                                  scalars = self._get_current_elem_numbers,
                                  #color = (0.254902,0.411765,0.882353)
                                  color = (0.15,0.85,0.45))
                                 
        
    mvp_elem_geo = Trait( MVPolyData )
    def _mvp_elem_geo_default(self):
        return MVPolyData( name = 'Geo node numbers', 
                               points = self._get_current_elem_coords,
                               lines = self._get_current_elem_lines,
                               #color = (0.254902,0.411765,0.882353))
                               color= (0,55,0,75,0.0))
        
    
    show_elem = Bool(True)
            
    def _get_current_elem_coords(self):
        return self.rt_domain._get_elem_coords( self.current_row )
    
    def _get_current_elem_numbers(self):
        return self.data[self.current_row]
    
    def _get_current_elem_lines(self):
        line_map = self.rt_domain.grid_cell_spec.cell_lines
        return line_map
    
    current_row = Int(-1)
    @on_trait_change('current_row')
    def redraw(self):
        if self.show_elem:
            self.mvp_elem_labels.redraw('label_scalars')
            self.mvp_elem_geo.redraw()
    
    elem_view = Instance( ElemView )
    def _elem_view_default(self):
        return ElemView()
    
    @on_trait_change('current_row')
    def _display_current_row(self):
        if self.current_row != -1:
            self.elem_view.elem_num = self.current_row
            elem_coords = self.rt_domain._get_elem_coords( self.current_row )
            self.elem_view.elem_coords = elem_coords




from traits.api import HasTraits
import numpy as np


class IStrainNorm3D(HasTraits):

    def get_eps_eq(self, eps_Emef, kappa_Em):
        raise NotImplementedError

    def get_deps_eq(self, eps_Emef):
        raise NotImplementedError


class Rankine(IStrainNorm3D):
    '''
    Computes principal strains and makes a norm of their positive part
    '''

    def get_eps_eq(self, eps_Emef, kappa_Em):

        eps_11 = eps_Emef[..., 0, 0]
        eps_22 = eps_Emef[..., 1, 1]
        eps_33 = eps_Emef[..., 2, 2]
        eps_12 = eps_Emef[..., 0, 1]
        eps_13 = eps_Emef[..., 0, 2]
        eps_23 = eps_Emef[..., 1, 2]

#         eps_eq_Em = (
#             0.5 * (eps_11 + eps_22 + eps_33) +
#             np.sqrt(((eps_11 - eps_22) / 2.0)**2.0 + eps_12**2.0)
#         )

        eps_eq_Eme = np.linalg.eig(eps_Emef)[0]
        eps_eq_Em = np.ndarray.max(eps_eq_Eme, axis=-1)

        e_Em = np.concatenate(
            (eps_eq_Em[..., None], kappa_Em[..., None]), axis=-1
        )
        eps_eq = np.max(e_Em, axis=-1)

        return eps_eq

    def get_deps_eq(self, eps_Emef):
        eps11 = eps_Emef[..., 0, 0]
        eps22 = eps_Emef[..., 1, 1]
        eps12 = eps_Emef[..., 0, 1]
        eps_11_22 = eps11 - eps22
        factor = 1. / (2. * np.sqrt(eps_11_22 * eps_11_22 +
                                    4.0 * eps12 * eps12))
        df_trial1 = factor * np.array([[eps11 - eps22, 4.0 * eps12],
                                       [4.0 * eps12, eps22 - eps11]])
        return (np.einsum('ab...->...ab', df_trial1) +
                0.5 * np.identity(2)[None, :, :])


if __name__ == '__main__':
    import sympy as sp

    eps_0, eps_f, kappa = sp.symbols(
        '\\varepsilon_0, \\varepsilon_\\mathrm{f}, \\kappa')
    f = 1 - (eps_0 / kappa) * sp.exp(- (kappa - eps_0) / (eps_f - eps_0))
    print(f)
    print(sp.diff(f, kappa))

    s_11, s_22, s_33, s_12, s_23, s_13 = sp.symbols(
        '\\varepsilon_{11},\\varepsilon_{22},\\varepsilon_{33},'
        '\\varepsilon_{12},\\varepsilon_{23},\\varepsilon_{13}')
    sig = sp.Matrix([[s_11, s_12, s_13],
                     [s_12, s_22, s_23],
                     [s_13, s_23, s_33]])
    sig_12 = sig.eigenvals()
    for sig_principal, item in list(sig_12.items()):
        print(sig_principal)
        sig_diff_11 = sp.diff(sig_principal, s_11)
        sig_diff_22 = sp.diff(sig_principal, s_22)
        sig_diff_33 = sp.diff(sig_principal, s_33)
        sig_diff_12 = sp.diff(sig_principal, s_12)
        sig_diff_23 = sp.diff(sig_principal, s_23)
        sig_diff_13 = sp.diff(sig_principal, s_13)
        print('11', sig_diff_11)
        print('12', sig_diff_12)
        print('22', sig_diff_22)




from ibvpy.tfunction import \
    LoadingScenario
from ibvpy.bcond import BCDof
from ibvpy.sim.sim_base import Simulator
from ibvpy.xmodel.xdomain_point import XDomainSinglePoint
from .mats3D import MATS3DDesmorat

class MATSExplore(Simulator):
    '''
    Simulate the loading histories of a material point in 2D space.
    '''

    node_name = 'Composite tensile test'

    def _bc_default(self):
        return [BCDof(
            var='u', dof=0, value=-0.001,
            time_function=LoadingScenario()
        )]

    def _model_default(self):
        return MATS3DDesmorat()

    def _xdomain_default(self):
        return XDomainSinglePoint()





'''
Created on Mar 25, 2020

@author: rch
'''

import numpy as np
import traits.api as tr


class ILinAlgSys(tr.Interface):
    pass


class LinAlgSolve(tr.HasStrictTraits):
    '''Example of a linear equation system

    Solve the system of equations of the form

    :math
        A u = b_0 - b_t

    To define the boundary conditions prescribing the 
    values on the left-hand-side a sequential
    inclusion of the condition is managed by this class.
    Thus, the registration of the condition is performed
    first to register the variables affected by the 
    problem.
    '''
    o_I = tr.Array(np.int_)

    n_dofs = tr.Property(depends_on='dof_map')

    @tr.cached_property
    def _get_n_dofs(self):
        return np.max(self.o_I) + 1

    A = tr.Array(np.float64)

    R_t = tr.Property(tr.Array(np.float64), depends_on='dof_map')

    @tr.cached_property
    def _get_R_t(self):
        return np.zeros((self.n_dofs,), np.float64)

    R_0 = tr.Property(tr.Array(np.float64), depends_on='dof_map')

    @tr.cached_property
    def _get_R_0(self):
        return np.zeros((self.n_dofs,), np.float64)

    # The dependency graph

    constraints = tr.List
    '''Constraints are stored as a list of arrays.
    '''

    def register_constraint(self, a_U, U_a=0, tf=lambda t: t,
                            b_alpha=[], alpha_b=[]):
        '''Register the prescribed values of u as linear 
        transformations - mappings
        '''
        self.constraints.append((a_U, U_a, tf, b_alpha, alpha_b))

    t_n1 = tr.Float(1.0)

    U_arr = tr.Property(depends_on='t_n1')

    @tr.cached_property
    def _get_U_arr(self):
        '''Get the value of the constraint for the current step.
        '''
        carr = np.array(self.constraints)
        U_arr = carr[:, 1]
        tfarr = carr[:, 2]
        val_arr = np.array(
            [tf(u)
             for tf, u in zip(tfarr, U_arr)], dtype=np.float64)
        return val_arr

    def apply_constraints(self):
        carr = np.array(self.constraints)
        a_arr = np.array(carr[:, 0], dtype=np.int_)
        A_aa = self.A[a_arr, a_arr]
        self.A[a_arr, :] = 0
        self.A[:, a_arr] = 0
        self.A[a_arr, a_arr] = A_aa
        self.R_0[:] = np.einsum('i,ij->j', self.U_arr, self.A[a_arr, :])

    def solve(self):
        '''Solve the system
        '''
        return np.linalg.solve(self.A, self.R_0 - self.R_t)


if __name__ == '__main__':
    las = LinAlgSolve(
        A=[[1, 2, 3], [2, 4, 1], [3, 1, 5]],
        o_I=[0, 1, 2],
    )
    B = np.copy(las.A)
    B[(1, 2), (1, 2)]

    # Todo
    # Right hand side
    # Bundled calls to a sparse format matrix
    # Sparse matrix
    # How to merge the protocol with the compressed row
    # Define the time functions generally as piecewise linear

    las.register_constraint(0, 1)
    las.register_constraint(1, 1)
    las.apply_constraints()
    print(las.constraints)
    print(las.A)
    print(las.solve())




from functools import reduce

from numpy import \
    array, mgrid, c_, \
    arange, zeros, multiply, sort, index_exp, add, \
    frompyfunc
from traits.api import \
    Array, Property, cached_property, \
    Instance, Trait, Button, on_trait_change, Tuple, \
    Int, Float, provides, Delegate, Callable, HasTraits

from .cell_array import CellView, CellArray, ICellArraySource
from .cell_grid_slice import CellGridSlice
from .cell_spec import CellSpec, GridCell


#--------------------------------------------------------------------------
# CellGrid
#--------------------------------------------------------------------------
@provides(ICellArraySource)
class CellGrid(HasTraits):
    '''
    Manage an array of cells defined within a structured grid.

    The distinction between the coordinate information supplied 
    in the arrays is done using the following naming convention:

    point - geometric points within the regular grid

    node - point specified in the grid_cell specification

    vertex - node with topological function (corner nodes)

    base_node - the first node of an element

    point_idx_grid ... enumeration of points within the point_grid

    cell_idx_grid ... enumeration of cells in the grid maps the 
                      topological index to the flattened index
                      ( ix, iy, iz ) -> ii

    cell_node_map ... array mapping the cell idx to the  point idx 
                      representing a node of the cell

    '''

    # Everything depends on the grid_cell_specification
    # defining the distribution of nodes within the cell.
    #
    grid_cell_spec = Instance(CellSpec)

    def _grid_cell_spec_default(self):
        return CellSpec()

    n_dims = Delegate('grid_cell_spec')

    # Grid cell template - gets repeated according to the
    # within the grid geometry specification
    #
    grid_cell = Property(depends_on='grid_cell_spec.+')

    @cached_property
    def _get_grid_cell(self):
        return GridCell(grid_cell_spec=self.grid_cell_spec)

    # Grid geometry specification
    #
    coord_min = Array(Float, value=[0., 0., 0.])
    coord_max = Array(Float, value=[1., 1., 1.])

    # Remark[rch]:
    # beware - the Int type is not regarded as a normal int
    # within an array and must be first converted to int array
    #
    # Had we defined int as the dtype of an array, there would
    # be errors during editing
    #
    shape = Array(Int, value=[1, 1, 1])

    # Derived specifier for element grid shape
    # It converts the Int array to int so that it can be
    # used by general numpy operators
    #
    cell_idx_grid_shape = Property(Array(int), depends_on='shape')

    @cached_property
    def _get_cell_idx_grid_shape(self):
        return array(self.shape, dtype=int)

    cell_idx_grid_size = Property(Int, depends_on='shape')

    @cached_property
    def _get_cell_idx_grid_size(self):
        return reduce(lambda x, y: x * y, self.cell_idx_grid_shape)

    # Grid with the enumeration of the cells respecting
    # the dimensionality. This grid is used to implement
    # the mapping between the cells and nodes.
    #
    cell_idx_grid = Property(Array(int), depends_on='shape')

    @cached_property
    def _get_cell_idx_grid(self):
        return arange(self.cell_idx_grid_size).reshape(self.cell_idx_grid_shape)

    def __getitem__(self, idx):
        '''High level access and slicing to the cells within the grid.

        The return value is a tuple with 
        1. array of cell indices
        2. array of nodes for each element
        3. array of coordinates for each node.
        '''
        return CellGridSlice(cell_grid=self, grid_slice=idx)

    #-------------------------------------------------------------------------
    # Shape and size characteristics used for both the idx_grid and point_grid
    #-------------------------------------------------------------------------
    def _get_point_grid_shape(self):
        '''Get the grid shape for the full index and point grids.

        This is the background grid. Some of the nodes can be unused by the
        cells depending on the specification in the grid_cell_spec.
        '''
        cell_shape = self.grid_cell_spec.get_cell_shape().astype(int)
        cell_idx_grid_shape = self.cell_idx_grid_shape
        return multiply(cell_shape - 1, cell_idx_grid_shape) + 1

    point_grid_size = Property(depends_on='shape,grid_cell_spec.node_coords')

    def _get_point_grid_size(self):
        '''Get the size of the full index and point grids
        '''
        shape = self._get_point_grid_shape()
        return reduce(lambda i, j: i * j, shape)

    #-------------------------------------------------------------------------
    # point_idx_grid - shaping and slicing methods for construction and orientation
    # point_idx_grid represents the enumeration of the nodes of the cell grid. It
    # serves for constructing the mappings between cells and nodes.
    #-------------------------------------------------------------------------
    def _get_point_idx_grid_slices(self):
        '''Get slices defining the index grid in a format suitable 
        for mgrid generation.
        '''
        subcell_shape = self.grid_cell_spec.get_cell_shape() - 1
        cell_idx_grid_shape = self.cell_idx_grid_shape
        return tuple([slice(0, c * g + 1)
                      for c, g in zip(subcell_shape, cell_idx_grid_shape)])

    point_idx_grid = Property(
        Array, depends_on='shape,grid_cell_spec.node_coords')

    @cached_property
    def _get_point_idx_grid(self):
        '''Get the index numbers of the points within the grid
        '''
        return arange(self.point_grid_size).reshape(self._get_point_grid_shape())

    #-------------------------------------------------------------------------
    # Unit cell enumeration - used as a template for enumerations in 3D
    #-------------------------------------------------------------------------
    idx_cell_slices = Property(Tuple)

    def _get_idx_cell_slices(self):
        '''Get slices extracting the first cell from the point index grid
        '''
        cell_shape = self.grid_cell_spec.get_cell_shape()
        return tuple([slice(0, c)
                      for c in cell_shape])

    idx_cell = Property(Array)

    def _get_idx_cell(self):
        '''Get the node map within a cell of a 1-3 dimensional grid

        The enumeration of nodes within a single cell is managed by the
        self.grid_cell. This must be adapted to the global enumeration of the grid.

        The innermost index runs over the z-axis. Thus, the index of the points
        on the z axis is [0,1,...,shape[2]-1]. The following node at the y axis has 
        the number [shape[2], shape[2]+1, ..., shape[2]*2].
        '''
        return self.point_idx_grid[self.idx_cell_slices]

    def get_cell_idx(self, offset):
        '''Get the address of the cell within the cell grid.
        '''
        idx_tuple = zeros((self.n_dims,), dtype='int_')
        roof = offset
        for i, n in enumerate(self.shape[-1:0:-1]):
            idx = roof / (n)
            roof -= idx * n
            idx_tuple[i] = idx
        idx_tuple[self.n_dims - 1] = roof
        return tuple(idx_tuple)

    def get_cell_offset(self, idx_tuple):
        '''Get the index of the cell within the flattened list.
        '''
        return self.cell_idx_grid[idx_tuple]

    #-------------------------------------------------------------------------
    # vertex slices
    #-------------------------------------------------------------------------
    vertex_slices = Property

    def _get_vertex_slices(self):
        cell_shape = self.grid_cell_spec.get_cell_shape()
        return tuple([slice(0, None, c - 1)
                      for c in cell_shape])

    #-------------------------------------------------------------------------
    # point_grid - shaping and slicing methods for construction and orientation
    #-------------------------------------------------------------------------
    point_x_grid_slices = Property

    def _get_point_x_grid_slices(self):
        '''Get the slices to be used for the mgrid tool 
        to generate the point grid.
        '''
        ndims = self.n_dims
        shape = self._get_point_grid_shape()
        return tuple([slice(float(self.coord_min[i]),
                            float(self.coord_max[i]),
                            complex(0, shape[i]))
                      for i in range(ndims)])

    #-------------------------------------------------------------------------
    # Geometry transformation
    #-------------------------------------------------------------------------
    geo_transform = Callable

    geo_transform_vfn = Property

    def _get_geo_transform_vfn(self):
        vfn_shell_stb = frompyfunc(self.geo_transform, 2, 3)

    #-------------------------------------------------------------------------
    # Point coordinates x - is parametric, X - is global
    #-------------------------------------------------------------------------
    point_x_grid = Property(
        depends_on='grid_cell_spec.+,shape,coord_min,coord_max')

    @cached_property
    def _get_point_x_grid(self):
        '''
        Construct the point grid underlying the mesh grid structure.
        '''
        return mgrid[self._get_point_x_grid_slices()]

    point_X_grid = Property

    def _get_point_X_grid(self):
        '''
        Construct the point grid underlying the mesh grid structure.
        '''
        x_dim_shape = self.point_x_grid.shape[1:]
        return array([self.point_X_arr[:, i].reshape(x_dim_shape)
                      for i in range(self.n_dims)], dtype='float64')

    point_x_arr = Property

    def _get_point_x_arr(self):
        return c_[tuple([x.flatten() for x in self.point_x_grid])]

    point_X_arr = Property(
        depends_on='grid_cell_spec,shape,coord_min,coord_max, geo_transform')

    @cached_property
    def _get_point_X_arr(self):
        '''Get the (n,3) array with point coordinates.
        '''
        # If the geo transform has been set, perform the mapping
        #
        if self.geo_transform:
            return self.geo_transform(self.point_x_arr)
        else:
            return self.point_x_arr

    #-------------------------------------------------------------------------
    # Vertex manipulations
    #-------------------------------------------------------------------------
    vertex_idx_grid = Property

    def _get_vertex_idx_grid(self):
        '''
        Construct the base node grid. Base node has the lowest number within a cell.
        All relevant cell_node numbers can be derived just be adding an array
        of relative node offsets within the cell to the base node.
        (see the method get_cell_nodes( cell_num)  below)  
        '''
        # get the cell shape - number of cell points without the next base
        # point
        subcell_shape = self.grid_cell_spec.get_cell_shape().astype(int) - 1
        # get the element grid shape (number of elements in each dimension)
        cell_idx_grid_shape = self.cell_idx_grid_shape

        # The following code determines the offsets between two neighbouring nodes
        # along each axis. It loops over the axes so that also 1- and 2- dimensional
        # grids are included. For 3D grid - the following code shows what's happening
        #
        # 1) get the index offset between two neighboring points on the z-axis
        #
        # z_offset = subcell_shape[2]
        #
        # 2) get the index offset between two neighboring points on the y-axis
        #
        # y_offset = ( cell_idx_grid_shape[2] * subcell_shape[2] + 1) * subcell_shape[1]
        #
        # 3) get the index offset between two neighboring points on the x-axis
        #
        # x_offset = ( cell_idx_grid_shape[2] * subcell_shape[2] + 1 ) * \
        #            ( cell_idx_grid_shape[1] * subcell_shape[1] + 1 ) * \
        #            subcell_shape[0]

        offsets = zeros(self.n_dims, dtype='int_')

        for i in range(self.n_dims - 1, -1, -1):
            offsets[i] = subcell_shape[i]
            for j in range(i + 1, self.n_dims):
                offsets[i] *= (cell_idx_grid_shape[j] * subcell_shape[j] + 1)

        # grid shape (shape + 1)
        gshape = cell_idx_grid_shape + 1

        # Determine the offsets of all base nodes on each axis by multiplying
        # the respective offset with a point enumeration on that axis
        #
        # In 3D corresponds to the following
        #
        #        xi_offsets = x_offset * arange( gshape[0] )
        #        yi_offsets = y_offset * arange( gshape[1] )
        #        zi_offsets = z_offset * arange( gshape[2] )

        all_offsets = [offsets[n] * arange(gshape[n])
                       for n in range(self.n_dims)]

        # Construct the broadcastable slices used for the construction of the
        # base node index grid. In 3D this corresponds to the following:
        #
        # Expand the dimension of offsets and sum them up. using broadcasting
        # this generates the grid of the base nodes (the last node is cut away
        # as it does not hold any element
        #        idx_grid = xi_offsets[:-1,None,None] + \
        #                   yi_offsets[None,:-1,None] + \
        #                   zi_offsets[None,None,:-1]

        slices = []
        for i in range(self.n_dims):
            s = [None for j in range(self.n_dims)]
            s[i] = slice(None)
            slices.append(tuple(s))

        vertex_offsets = [all_offsets[i][slices[i]]
                          for i in range(self.n_dims)]
        vertex_idx_grid = reduce(add, vertex_offsets)

        # return the indexes of the vertex nodes
        return vertex_idx_grid

    vertex_idx_arr = Property

    def _get_vertex_idx_arr(self):
        '''Get the index of nodes at the vertex of the cells
        of the cells. The result is a sorted flat array. The base node 
        position within the returned array defines the index of the cell. 
        '''
        vertex_idx_grid = self.vertex_idx_grid
        return sort(vertex_idx_grid.flatten())

    vertex_x_grid = Property

    def _get_vertex_x_grid(self):
        return array([x_grid[self.vertex_slices] for x_grid in self.point_x_grid])

    vertex_X_grid = Property

    def _get_vertex_X_grid(self):
        return array([X_grid[self.vertex_slices] for X_grid in self.point_X_grid])

    vertex_x_arr = Property

    def _get_vertex_x_arr(self):
        return c_[tuple([x.flatten() for x in self.vertex_x_grid])]

    vertex_X_arr = Property

    def _get_vertex_X_arr(self):
        return c_[tuple([X.flatten() for X in self.vertex_X_grid])]

    #-------------------------------------------------------------------------
    # Cell manipulations
    #-------------------------------------------------------------------------
    base_nodes = Property

    def _get_base_nodes(self):
        '''Get the index of nodes that are the bottom left front vertexs
        of the cells. The result is a sorted flat array. The base node 
        position within the returned array defines the index of the cell. 
        '''
        vertex_idx_grid = self.vertex_idx_grid
        cutoff_last = [slice(0, -1) for i in range(self.n_dims)]
        base_node_grid = vertex_idx_grid[tuple(cutoff_last)]
        return sort(base_node_grid.flatten())

    cell_node_map = Property(depends_on='shape,grid_cell_spec.+')

    @cached_property
    def _get_cell_node_map(self):
        '''
        Construct an array with the mapping between elements and nodes. 
        Returns the dof for [ cell_idx, node, dim ]
        '''
        idx_cell = self.idx_cell
        node_map = idx_cell.flatten()[self.grid_cell.node_map]
        base_nodes = self.base_nodes

        # Use broadcasting to construct the node map for all elements
        #
        cell_node_map = base_nodes[:, None] + node_map[None, :]
        return cell_node_map

    cell_grid_node_map = Property(depends_on='shape,grid_cell_spec.+')

    @cached_property
    def _get_cell_grid_node_map(self):
        '''
        Return the dof for [ cell_x, cell_y, node, dim ]
        where 
        - cell_x - is the cell index in the first dimension
        - cell_y - is the cell index in the second dimension
        - node - is the node index within the cell
        - dim - is the index of the dof within the node
        '''
        new_shape = tuple(self.shape) + self.cell_node_map.shape[1:]
        return self.cell_node_map.reshape(new_shape)

    def get_cell_point_x_arr(self, cell_idx):
        '''Return the node coordinates included in the cell cell_idx. 
        '''
        iexp = index_exp[self.cell_node_map[cell_idx]]
        return self.point_x_arr[iexp]

    def get_cell_point_X_arr(self, cell_idx):
        '''Return the node coordinates included in the cell cell_idx. 
        '''
        iexp = index_exp[self.cell_node_map[cell_idx]]
        return self.point_X_arr[iexp]

    #-------------------------------------------------------------------------
    # @todo - candidate for deletion - the slicing operator [] is doing this more generally
    #-------------------------------------------------------------------------
    boundary_slices = Property(depends_on='grid_cell_spec.+')

    @cached_property
    def _get_boundary_slices(self):
        '''Get the slices to get the boundary nodes.
        '''
        # slices must correspond to the dimensions
        slices = []
        for i in range(self.n_dims):
            s_low = [slice(None) for j in range(self.n_dims)]
            s_low[i] = 0
            s_high = [slice(None) for j in range(self.n_dims)]
            s_high[i] = -1
            slices.append(s_low)
            slices.append(s_high)
        return slices

    #--------------------------------------------------------------------------
    # Wrappers exporting the grid date to mayavi pipelines
    #--------------------------------------------------------------------------\

    def get_cell_mvpoints(self, cell_idx):
        '''Return the node coordinates included in the cell cell_idx. 
        In case that the grid is in reduced dimensions - blow it up with zeros.
        '''
        points = self.get_cell_point_X_arr(cell_idx)

        # augment the points to be 3D
        if self.n_dims < 3:
            mvpoints = zeros((points.shape[0], 3), dtype='float64')
            mvpoints[:, :self.n_dims] = points
            return mvpoints
        else:
            return points

    def _get_mvpoints_grid_shape(self):
        '''Shape of point grid in 3D.
        The information is needed by the mayavi pipeline to use the
        StructuredGrid source - tvtk class.
        '''
        shape = self._get_point_grid_shape()
        return tuple(list(shape) + [1 for n in range(3 - len(shape))])

    def _get_mvpoints(self, swap=True):
        '''Get the points in with deepest index along the x axes.
        This format is required when putting the point data to the 
        vtk structured grid.
        '''
        point_X_grid = self.point_X_grid
        if swap == True:
            point_X_grid = point_X_grid.swapaxes(1, self.n_dims)

        point_X_arr = c_[tuple([X.flatten() for X in point_X_grid])]

        # augment the points to be 3D
        if self.n_dims < 3:
            mv_points = zeros((point_X_arr.shape[0], 3), dtype='float64')
            mv_points[:, :self.n_dims] = point_X_arr
            return mv_points
        else:
            return point_X_arr

    #-----------------------------------------------------------------
    # Visualization-related methods
    #-----------------------------------------------------------------

    # mvp_point_grid = Trait(MVStructuredGrid)
    #
    # def _mvp_point_grid_default(self):
    #     return MVStructuredGrid(name='Point grid',
    #                             dims=self._get_mvpoints_grid_shape,
    #                             points=self._get_mvpoints)

    refresh_button = Button('Draw')

    @on_trait_change('refresh_button')
    def redraw(self):
        '''Redraw the point grid.
        '''
        self.mvp_point_grid.redraw()

    cell_array = Button('Browse cell array')

    def _cell_array_fired(self):
        cell_array = self.cell_node_map
        self.show_array = CellArray(data=cell_array,
                                    cell_view=CellView(cell_grid=self))
        self.show_array.configure_traits(kind='live')




'''
Created on Dec 17, 2018

@author: rch
'''
import unittest

import numpy as np
from simulator.sim_controler import Model
import traits.api as tr


class SinusModel(Model):

    #=========================================================================
    # Model implementation - F and Fprime
    #=========================================================================

    def _get_F(self):
        return np.sin(self.U_k)

    def _get_d_F_U(self):
        return [np.cos(self.U_k)]

    #=========================================================================
    # Derived variables
    #=========================================================================
    G = tr.Property

    def _get_G(self):
        return np.arctan(self.U_k)


class Test(unittest.TestCase):

    def setUp(self):
        self.m = SinusModel()

    def test_model_init(self):
        m = self.m
        self.assertTrue(m == m.tstep.model, 'Model initialization success')
        self.assertTrue(m.tstep == m.tloop.tstep,
                        'Model initialization success')
        self.assertTrue(m.hist == m.sim.hist,
                        'Model initialization success')

    def test_tloop_eval(self):
        '''Run the model evaluation.

        For convenience, the model should incorporate
        the boundary conditions as well.
        '''
        m = self.m
        m.tloop.init()
        m.tloop.eval()
        self.assertAlmostEqual(m.U_k[0], 1.55702096)

    def test_bc(self):
        pass


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()




from .vmats1D5_d import MATS1D5D
from .vmats1D5_bondslip1D import MATS1D5BondSlipMultiLinear, \
    MATS1D5BondSlipD, MATS1D5BondSlipEP
from .vmats1D5_bondslip1D_trilinear import MATS1D5BondSlipTriLinear


from traits.api import \
    Instance, Int, Property, Array, cached_property, Instance

from numpy import \
     zeros, dot, hstack, vstack, identity, cross, transpose, tensordot, outer

from scipy.linalg import \
     inv, norm

from ibvpy.fets.fets_eval import FETSEval
from ibvpy.fets.fets3D.fets3D8h20u import FETS3D8H20U

# import the 3D elastic matrix in order
from ibvpy.tmodel.mats3D.mats3D_elastic.mats3D_elastic import \
    MATS3DElastic

from ibvpy.tmodel.mats2D.mats2D_tensor import \
    map2d_eps_mtx_to_eng, map2d_sig_eng_to_mtx

from ibvpy.tmodel.mats3D.mats3D_tensor import \
    map3d_eps_eng_to_mtx, map3d_sig_eng_to_mtx, map3d_sig_mtx_to_eng, map3d_eps_mtx_to_eng, map3d_tns2_to_tns4, map3d_tns4_to_tns2, map3d_eps_mtx_to_eng

# from coordinate_transformations import \
#    get_Trans_strain_from_glb_to_loc, get_Trans_stress_from_loc_to_glb

# from i_fets_eval import IFETSEval

#-----------------------------------------------------------------------------------
# FETS2D58H20U -  degenerated subparametric 3D volume element using a 2D material model
#                 quadratic (serendipity)
#-----------------------------------------------------------------------------------

class FETS2D58H20U(FETS3D8H20U):

    # @todo: the calculation of the direction cosines
    #        should be cached (similar to J_mtx and B_mtx)

    def get_dircos_mtx(self, sctx):
        '''returns the direction cosine at the integration point.
        The three column vectors v1, v2, v3 of the direction cosine
        matrix build an orthonormal system and can be used to rotate
        the 3D strains into the local coordinate system.
        The vector v3 stands perpendicular to the plane defined by the
        curvilinear coordinates r and s. Note that for the orientation
        of the local in-plane coordinates different choices exist.'''

        # get the Jacobian matrix
        r_pnt = sctx.r_pnt
        X_mtx = sctx.X
        J_mtx = self.get_J_mtx(r_pnt, X_mtx)

        # the first two rows of the Jacobi matrix are tangential
        # to the mid-surface at the integration point.
        # (cf. Zienkiewicz Vol.2, Eq.(15.19))
        j1 = J_mtx[0, :]
        j2 = J_mtx[1, :]

        # v3 stands perdendicular to the surface
        v3 = cross(j1, j2)

        # One possible choice for the direction of the local x-axis
        # is to define it such that it coincides with the orientation
        # of the first curvilinear coordinate r (cf. FE-Skript)
        v1 = j1

        # The local y-axis is obtained as the cross product of the
        # local x and z direction
        v2 = cross(v3, v1)

        # normalize the vectors to obtain an ortho-normal basis
        v1_ = v1 / norm(v1)
        v2_ = v2 / norm(v2)
        v3_ = v3 / norm(v3)

        return transpose(vstack((v1_, v2_, v3_)))

    def get_Trans_strain_from_glb_to_loc(self, dircos_mtx):
        '''Transformation matrix in 3d shape = (6,6).
        (eps_mtx_loc = T * eps_mtx_glb * T^t)
        Multiply the global strain vector 'eps_vct_glb' (in engineering notation) with the returned
        transformation matrix 'Trans_strain_from_glb_to_loc_mtx' in order to obtain the loacal strain vector
        (in engineering notation) in the local coordinate system (x',y',z'), e.g. in-plan shell directions.
        The ordering of the strain vector in engineering notation corresponds to the VOIGT-notation:
        eps_glb = [eps_xx, eps_yy, eps_zz, eps_yz, eps_zx, eps_xz].
        The parameter 'dircos_mtx' is the matrix of the direction cosines describing the relation between
        the global and local coordinate system.
        (the returned transformation matrix corresponds to Zienkiewicz Eq.(6.24) but used VOIGT-notation instead)
        '''
        t = dircos_mtx  # shape=(3,3)
        T = zeros((6, 6))  # shape=(6,6)
        T[0, 0] = t[0, 0] * t[0, 0]
        T[0, 1] = t[0, 1] * t[0, 1]
        T[0, 2] = t[0, 2] * t[0, 2]
        T[0, 3] = t[0, 1] * t[0, 2]
        T[0, 4] = t[0, 0] * t[0, 2]
        T[0, 5] = t[0, 0] * t[0, 1]
        T[1, 0] = t[1, 0] * t[1, 0]
        T[1, 1] = t[1, 1] * t[1, 1]
        T[1, 2] = t[1, 2] * t[1, 2]
        T[1, 3] = t[1, 1] * t[1, 2]
        T[1, 4] = t[1, 0] * t[1, 2]
        T[1, 5] = t[1, 0] * t[1, 1]
        T[2, 0] = t[2, 0] * t[2, 0]
        T[2, 1] = t[2, 1] * t[2, 1]
        T[2, 2] = t[2, 2] * t[2, 2]
        T[2, 3] = t[2, 1] * t[2, 2]
        T[2, 4] = t[2, 0] * t[2, 2]
        T[2, 5] = t[2, 0] * t[2, 1]
        T[3, 0] = 2.0 * t[2, 0] * t[1, 0]
        T[3, 1] = 2.0 * t[2, 1] * t[1, 1]
        T[3, 2] = 2.0 * t[2, 2] * t[1, 2]
        T[3, 3] = t[2, 1] * t[1, 2] + t[2, 2] * t[1, 1]
        T[3, 4] = t[2, 2] * t[1, 0] + t[2, 0] * t[1, 2]
        T[3, 5] = t[2, 0] * t[1, 1] + t[2, 1] * t[1, 0]
        T[4, 0] = 2.0 * t[2, 0] * t[0, 0]
        T[4, 1] = 2.0 * t[2, 1] * t[0, 1]
        T[4, 2] = 2.0 * t[2, 2] * t[0, 2]
        T[4, 3] = t[2, 1] * t[0, 2] + t[2, 2] * t[0, 1]
        T[4, 4] = t[2, 2] * t[0, 0] + t[2, 0] * t[0, 2]
        T[4, 5] = t[2, 0] * t[0, 1] + t[2, 1] * t[0, 0]
        T[5, 0] = 2.0 * t[1, 0] * t[0, 0]
        T[5, 1] = 2.0 * t[1, 1] * t[0, 1]
        T[5, 2] = 2.0 * t[1, 2] * t[0, 2]
        T[5, 3] = t[1, 1] * t[0, 2] + t[1, 2] * t[0, 1]
        T[5, 4] = t[1, 2] * t[0, 0] + t[1, 0] * t[0, 2]
        T[5, 5] = t[1, 0] * t[0, 1] + t[1, 1] * t[0, 0]
        Trans_strain_to_loc_mtx = T
        return Trans_strain_to_loc_mtx

    def get_Trans_stress_from_loc_to_glb(self, dircos_mtx):
        '''Transformation matrix in 3d shape = (6,6)
        (sig_mtx_glb = T^t * sig_mtx_loc * T)
        Multiply the local stress vector 'sig_vct_loc' (in engineering notation in the local coordinate
        system (x',y', z') with the returned transformation matrix 'Trans_stress_from_loc_to_glb_mtx' in
        order to obtain the global stress vector 'sig_vct_glb' (in engineering notation) in the global
        coordinate system (x,y,z). (cf. Zienkiewicz Eq.(6.23) multiplied with T_t*(...)*T, for T being an orthonormal basis)
        The ordering of the stress vector in engineering notation corresponds to the VOIGT-notation:
        sig_vct_glb = [sig_xx, sig_yy, sig_zz, sig_yz, sig_zx, sig_xz].
        The parameter 'dircos_mtx' is the matrix of the direction cosines describing the relation between
        the global and the local coordinate system.
        '''
        t = dircos_mtx  # shape=(3,3)
        T = zeros((6, 6))  # shape=(6,6)
        T[0, 0] = t[0, 0] * t[0, 0]
        T[0, 1] = t[1, 0] * t[1, 0]
        T[0, 2] = t[2, 0] * t[2, 0]
        T[0, 3] = 2.0 * t[1, 0] * t[2, 0]
        T[0, 4] = 2.0 * t[0, 0] * t[2, 0]
        T[0, 5] = 2.0 * t[0, 0] * t[1, 0]
        T[1, 0] = t[0, 1] * t[0, 1]
        T[1, 1] = t[1, 1] * t[1, 1]
        T[1, 2] = t[2, 1] * t[2, 1]
        T[1, 3] = 2.0 * t[1, 1] * t[2, 1]
        T[1, 4] = 2.0 * t[0, 1] * t[2, 1]
        T[1, 5] = 2.0 * t[0, 1] * t[1, 1]
        T[2, 0] = t[0, 2] * t[0, 2]
        T[2, 1] = t[1, 2] * t[1, 2]
        T[2, 2] = t[2, 2] * t[2, 2]
        T[2, 3] = 2.0 * t[1, 2] * t[2, 2]
        T[2, 4] = 2.0 * t[0, 2] * t[2, 2]
        T[2, 5] = 2.0 * t[0, 2] * t[1, 2]
        T[3, 0] = t[0, 2] * t[0, 1]
        T[3, 1] = t[1, 2] * t[1, 1]
        T[3, 2] = t[2, 2] * t[2, 1]
        T[3, 3] = t[1, 2] * t[2, 1] + t[2, 2] * t[1, 1]
        T[3, 4] = t[2, 2] * t[0, 1] + t[0, 2] * t[2, 1]
        T[3, 5] = t[0, 2] * t[1, 1] + t[1, 2] * t[0, 1]
        T[4, 0] = t[0, 2] * t[0, 0]
        T[4, 1] = t[1, 2] * t[1, 0]
        T[4, 2] = t[2, 2] * t[2, 0]
        T[4, 3] = t[1, 2] * t[2, 0] + t[2, 2] * t[1, 0]
        T[4, 4] = t[2, 2] * t[0, 0] + t[0, 2] * t[2, 0]
        T[4, 5] = t[0, 2] * t[1, 0] + t[1, 2] * t[0, 0]
        T[5, 0] = t[0, 1] * t[0, 0]
        T[5, 1] = t[1, 1] * t[1, 0]
        T[5, 2] = t[2, 1] * t[2, 0]
        T[5, 3] = t[1, 1] * t[2, 0] + t[2, 1] * t[1, 0]
        T[5, 4] = t[2, 1] * t[0, 0] + t[0, 1] * t[2, 0]
        T[5, 5] = t[0, 1] * t[1, 0] + t[1, 1] * t[0, 0]
        Trans_stress_from_loc_mtx = T
        return Trans_stress_from_loc_mtx

    def get_eps_eng(self, sctx, u):

        eps_eng3D = super(FETS2D58H20U, self).get_eps_eng(sctx, u)

        # get the direction cosine at the ip:
        dircos_mtx = self.get_dircos_mtx(sctx)

        # get the Transformation matrices for switching between the
        # local and global coordinate system in 3D case:
        T_strain_to_loc_mtx = self.get_Trans_strain_from_glb_to_loc(dircos_mtx)

        # transform strain from global to local coordinates
        #
        eps_eng3D_ = dot(T_strain_to_loc_mtx, eps_eng3D)

        return eps_eng3D_

    def get_mtrl_corr_pred(self, sctx, eps_eng3D, d_eps_eng3D, tn, tn1, eps_avg=None):
        '''
        Overload the 3D material-corrector-predictor in
        order to transform the strain and stress tensor
        into the in-plane so that oriented material model formulation
        can be applied.

        This is performed
        employing the direction cosines at each integration
        point (ip). The direction cosines establish the link
        between the global and local coordinate system (the
        reference to the local coordinate system is indicated
        by the underline character at the end of a variable
        name).

        After employing the material model the returned
        stiffness and stresses are transformed back to the global
        coordinate system.
        '''

        # get the direction cosine at the ip:
        dircos_mtx = self.get_dircos_mtx(sctx)

        # get the Transformation matrices for switching between the
        # local and global coordinate system in 3D case:
        T_stress_from_loc_mtx = self.get_Trans_stress_from_loc_to_glb(dircos_mtx)
        T_strain_to_loc_mtx = self.get_Trans_strain_from_glb_to_loc(dircos_mtx)

        # transform strain from global to local coordinates
        #
        eps_eng3D_ = dot(T_strain_to_loc_mtx, eps_eng3D)
        d_eps_eng3D_ = dot(T_strain_to_loc_mtx, d_eps_eng3D)

        # evaluate the mtrl-model
        #
        sig_eng3D_, D_mtx3D_ = self.mats_eval.get_corr_pred(sctx, eps_eng3D_, d_eps_eng3D_,
                                                             tn, tn1)

        #------------------------------------------------------------------------
        # get predictor (D_mtx)
        #------------------------------------------------------------------------

        # transform D_mtx3D from local to global coordinates
        #
        D_mtx3D = dot(T_stress_from_loc_mtx, dot(D_mtx3D_, T_strain_to_loc_mtx))

        #------------------------------------------------------------------------
        # get corrector (sig_eng)
        #------------------------------------------------------------------------

        # transform sig_mtx3D from local to global coordinates
        #
        sig_eng3D = dot(T_stress_from_loc_mtx, sig_eng3D_)

        return sig_eng3D, D_mtx3D




from traits.api import Interface


class IBCond(Interface):
    '''
    Interface of the boundary condition.
    '''

    def is_essential(self):
        '''
        Distinguish the essential and natural boundary conditions.

        This is needed to reorganize the system matrices and vectors
        to reflect the prescribed primary variables (displacements).
        '''

    def is_natural(self):
        '''
        Distinguish the essential and natural boundary conditions.

        This is needed to reorganize the system matrices and vectors
        to reflect the prescribed primary variables (displacements).
        '''

#    def get_dofs( self ):
#        '''
#        Return the list of affected DOFs.
#
#        This is needed to reorganize the system matrices and vectors
#        to reflect the prescribed primary variables (displacements).
#        '''

    def setup(self, sctx):
        '''
        Locate the spatial context.
        '''

    def apply_essential(self, K):
        '''
        Locate the spatial context.
        '''

    def apply(self, step_flag, sctx, K, R, t_n, t_n1):
        '''
        Locate the spatial context.
        '''



'''
Created on Aug 8, 2009

@author: rch
'''
import unittest
from ibvpy.api import \
    TStepper as TS, RTDofGraph, RTraceDomainListField, TLoop, \
    TLine, BCDofGroup, IBVPSolve as IS, DOTSEval, BCSlice

from ibvpy.mesh.fe_grid import FEGrid

from ibvpy.tmodel.mats3D.mats3D_elastic.mats3D_elastic import MATS3DElastic
from ibvpy.tmodel.mats2D.mats2D_elastic.mats2D_elastic import MATS2DElastic

# linear
from ibvpy.fets.fets3D.fets3D8h import FETS3D8H
from ibvpy.fets.fets2D5.fets2D58h import FETS2D58H

# quadratic serendipity
from ibvpy.fets.fets3D.fets3D8h20u import FETS3D8H20U
from ibvpy.fets.fets2D5.fets2D58h20u import FETS2D58H20U

# quadratic/linear hybird
from ibvpy.fets.fets3D.fets3D8h16u import FETS3D8H16U
from ibvpy.fets.fets2D5.fets2D58h16u import FETS2D58H16U

# cubic serendipity
from ibvpy.fets.fets3D.fets3D8h32u import FETS3D8H32U
from ibvpy.fets.fets2D5.fets2D58h32u import FETS2D58H32U

# cubic/linear hybird
from ibvpy.fets.fets3D.fets3D8h24u import FETS3D8H24U
from ibvpy.fets.fets2D5.fets2D58h24u import FETS2D58H24U


from ibvpy.util.simgrid import simgrid

from ibvpy.tmodel.mats2D5.mats2D5_cmdm.mats2D5_cmdm import \
    MATS2D5MicroplaneDamage

from ibvpy.tmodel.matsXD.matsXD_cmdm.matsXD_cmdm_phi_fn import \
    PhiFnStrainSoftening, PhiFnGeneral


class TestFETS2D5( unittest.TestCase ):


    def setUp( self ):
        # test elastic case: compare values of MATS2D5 with no damage (phi_fn = 1 (const.)).
        phi_fn = PhiFnGeneral()
        print(phi_fn.mfn.xdata)
        print(phi_fn.mfn.ydata)

        # linear elements
        self.fets_eval3D = FETS3D8H( mats_eval = MATS3DElastic( E = 34000., nu = 0.25 ) )
        self.fets_eval2D5 = FETS2D58H( mats_eval = MATS2D5MicroplaneDamage( E = 34000., nu = 0.25,
                                                                          model_version = "compliance",
                                                                          #model_version   = "stiffness",
                                                                          phi_fn = phi_fn ) )

#        # quadratic serendipity elements
#        self.fets_eval3D = FETS3D8H20U( mats_eval = MATS3DElastic( E = 34000., nu = 0.25 ) )
#        self.fets_eval2D5 = FETS2D58H20U( mats_eval = MATS2D5MicroplaneDamage( E = 34000., nu = 0.25,
#                                                                          phi_fn = phi_fn ) )

#        # quadratic/linear hybrid elements
#        self.fets_eval3D = FETS3D8H16U( mats_eval = MATS3DElastic( E = 34000., nu = 0.25 ) )
#        self.fets_eval2D5 = FETS2D58H16U( mats_eval = MATS2D5MicroplaneDamage( E = 34000., nu = 0.25,
#                                                                          phi_fn = phi_fn ) )

#        # cubic serendipity elements
#        self.fets_eval3D = FETS3D8H32U( mats_eval = MATS3DElastic( E = 34000., nu = 0.25 ) )
#        self.fets_eval2D5 = FETS2D58H32U( mats_eval = MATS2D5MicroplaneDamage( E = 34000., nu = 0.25,
#                                                                          phi_fn = phi_fn ) )

#        # cubic/linear hybrid elements
#        self.fets_eval3D = FETS3D8H24U( mats_eval = MATS3DElastic( E = 34000., nu = 0.25 ) )
#        self.fets_eval2D5 = FETS2D58H24U( mats_eval = MATS2D5MicroplaneDamage( E = 34000., nu = 0.25,
#                                                                          phi_fn = phi_fn ) )


    def test_uniform_loading( self ):

        support_slices = [
                          [ ( 0   , slice( None ), slice( None ), 0   , slice( None ), slice( None ) ), # yz plane  0
                            ( 0   , 0   , slice( None ), 0   , 0   , slice( None ) ), #  z-axis   1
                            ( 0   , 0   , 0, 0   , 0   , 0 )  #  origin   2
                          ],
                          [
                            ( 0   , 0   , 0   , 0   , 0   , 0 ), #  origin   0
                            ( slice( None ), 0   , slice( None ), slice( None ), 0   , slice( None ) ), # xz plane  1
                            ( slice( None ), 0   , 0   , slice( None ), 0   , 0 ), #  x-axis   2
                          ],
                          [
                            ( 0   , slice( None ), 0   , 0   , slice( None ), 0 ), #  y-axis   0
                            ( 0   , 0   , 0   , 0   , 0   , 0 ), #  origin   1
                            ( slice( None ), slice( None ), 0   , slice( None ), slice( None ), 0 ), # xz plane  2
                          ],
                          ]
        support_dirs = [[0], [1], [2]]

        loading_slices = [
                          ( -1  , slice( None ), slice( None ), -1  , slice( None ), slice( None ) ), # loading in x dir
                          ( slice( None ), -1  , slice( None ), slice( None ), -1  , slice( None ) ), # loading in y dir
                          ( slice( None ), slice( None ), -1  , slice( None ), slice( None ), -1 )   # loading in z dir
                        ]

        load_dirs = [0, 1, 2]
        load = 0.01
        vars = ['eps_app']
        for support_slice, loading_slice in zip( support_slices, loading_slices ):

            for load_dir in load_dirs:
                tl, u3D, fields3D, integs, g = simgrid( self.fets_eval3D, ( 3, 3, 3 ), ( 1, 1, 1 ),
                                           support_slice, support_dirs,
                                           loading_slice, load_dir,
                                           load, 1, vars )

                tl, u2D5, fields2D5, integs, g = simgrid( self.fets_eval2D5, ( 3, 3, 3 ), ( 1, 1, 1 ),
                                           support_slice, support_dirs,
                                           loading_slice, load_dir,
                                           load, 1, vars )

                for u1_, u2_ in zip( u3D.flatten(), u2D5.flatten() ):
                    self.assertAlmostEqual( u1_, u2_ )

    def tearDown( self ):
        pass


if __name__ == "__main__":
    unittest.main()




from ibvpy.mathkit.linalg import SysMtxArray, DenseMtx, SysMtxAssembly
from ibvpy.mathkit.tensor import EPS
from ibvpy.sim.i_xmodel import IXModel
from traits.api import \
    Property, Array, Int, HasStrictTraits, \
    cached_property, Tuple, provides, Instance, DelegatesTo, \
    Constant, Bool
from ibvpy.view.ui import BMCSLeafNode

import matplotlib.pyplot as plt
import numpy as np

class LatticeTessellation(BMCSLeafNode):
    '''Lattice tessellation
    '''
    name = 'Tessellation'
    node_name = 'tessellation'

    X_Ia = Array(np.float64,
                 MESH=True,
                 input=True,
                 unit=r'm',
                 symbol=r'$X_{Ia}$',
                 auto_set=False, enter_set=True,
                 desc='node coordinates')
    I_Li = Array(np.int_,
                 MESH=True,
                 input=True,
                 unit='-',
                 symbol='$I_{Li}$',
                 auto_set=False, enter_set=True,
                 desc='connectivity')

    dof_offset = Int(0)

    n_dofs = Property()

    def _get_n_dofs(self):
        return len(self.X_Ia) * 6

    vtk_expand_operator = Array(np.float64, value=np.identity(3))

    n_nodal_dofs = Constant(3)

    plot_backend = 'k3d'

    def update_plot(self, k3d_plot):
        k3d_plot


@provides(IXModel)
class XDomainLattice(BMCSLeafNode):

    hidden = Bool(False)

    #=========================================================================
    # Methods needed by XDomain to chain the subdomains
    #=========================================================================
    dof_offset = DelegatesTo('mesh')

    def set_next(self, next_):
        self.mesh.next_grid = next_.mesh

    def set_prev(self, prev):
        self.mesh.prev_grid = prev.mesh

    mesh = Instance(LatticeTessellation)

    X_Ia = DelegatesTo('mesh')
    I_Li = DelegatesTo('mesh')
    n_dofs = DelegatesTo('mesh')
    vtk_expand_operator = DelegatesTo('mesh')
    n_nodal_dofs = DelegatesTo('mesh')

    U_var_shape = Property(Int)

    def _get_U_var_shape(self):
        return len(self.X_Ia.shape) * 3 * 2

    state_var_shape = Property(Tuple)

    def _get_state_var_shape(self):
        return (len(self.I_Li),)

    nT_Lba = Property(depends_on='MESH')

    @cached_property
    def _get_nT_Lba(self):
        X_Lia = self.X_Ia[self.I_Li]
        Tu_La = X_Lia[..., 1, :] - X_Lia[..., 0, :]
        I = np.fabs(Tu_La[..., 0]) > np.fabs(Tu_La[..., 2])
        Tvv_La = np.c_[0 * Tu_La[..., 0], -Tu_La[..., 2], Tu_La[..., 1]]
        Tvv_La[I, :] = np.c_[-Tu_La[I, 1], Tu_La[I, 0], 0 * Tu_La[I, 0]]
        Tw_La = np.einsum('abc,...a,...b->...c', EPS, Tu_La, Tvv_La)
        Tv_La = np.einsum('abc,...a,...b->...c', EPS, Tw_La, Tu_La)
        T_Lba = np.einsum('...bla->...lba', np.array([Tu_La, Tv_La, Tw_La]))
        norm_T_lb = 1. / np.sqrt(np.einsum(
            '...lba,...lba->...lb', T_Lba, T_Lba)
        )
        nT_Lba = np.einsum('...lb,...lba->...lba', norm_T_lb, T_Lba)
        return nT_Lba

    B_Lipac = Property(depends_on='MESH')

    @cached_property
    def _get_B_Lipac(self):
        DELTA2 = np.identity(2)
        X_Lia = self.X_Ia[self.I_Li]
        Xm_La = np.einsum('ii,Lia->La', DELTA2, X_Lia) / 2
        Xm_Lia = Xm_La[..., np.newaxis, :]
        dXm_Lia = Xm_Lia - X_Lia
        switch_sign = np.array([-1, 1], dtype=np.float64)
        S_Liac = np.einsum('i,ii,...Lac->...Liac',
                           switch_sign, DELTA2, self.nT_Lba)
        return np.einsum('pLiac->Lipac', np.array(
            [S_Liac, np.einsum('Liab,bcd,Lid->Liac',
                               S_Liac, EPS, dXm_Lia)]
        ))

    B_Lipabjqcd = Property(depends_on='MESH,GEO,CS')

    @cached_property
    def _get_B_Lipabjqcd(self):
        return np.einsum(
            '...Lipab,...Ljqcd->...Lipabjqcd',
            self.B_Lipac, self.B_Lipac
        )

    o_Ipa = Property(depends_on='MESH')

    @cached_property
    def _get_o_Ipa(self):
        n_I, _ = self.X_Ia.shape
        n_a = 3
        n_p = 2
        return np.arange(n_I * n_p * n_a, dtype=np.int_).reshape(-1, n_p, n_a)

    o_Lipa = Property(depends_on='MESH,GEO,CS,FE')

    @cached_property
    def _get_o_Lipa(self):
        I_Li = self.I_Li
        return self.o_Ipa[I_Li]

    def map_U_to_field(self, U):
        U_Lipa = U[self.o_Lipa]
        return np.einsum('Lipac,Lipc->La', self.B_Lipac, U_Lipa)

    def map_field_to_F(self, sig_La):
        n_p, _, n_i, _, n_c = self.B_Lipac.shape
        f_Lipa = np.einsum('Lipac,La->Lipc', self.B_Lipac, sig_La)
        f_Li = f_Lipa.reshape(-1, n_i * n_p * n_c)
        o_L = self.o_Lipa.reshape(-1, n_i * n_p * n_c)
        return o_L.flatten(), f_Li.flatten()

    def map_field_to_K(self, D_Lbc):
        K_Lipbjqd = np.einsum(
            'Lipabjqcd,Lac->Lipbjqd', self.B_Lipabjqcd, D_Lbc
        )
        _, n_i, n_p, n_c, _, _, _ = K_Lipbjqd.shape
        K_Lij = K_Lipbjqd.reshape(-1, n_i * n_p * n_c, n_i * n_p * n_c)
        o_Li = self.o_Lipa.reshape(-1, n_i * n_p * n_c)
        return SysMtxArray(mtx_arr=K_Lij, dof_map_arr=o_Li)

    Xm_Lia = Property

    def _get_Xm_Lia(self):
        X_Lia = self.X_Ia[self.I_Li]
        DELTA2 = np.identity(2)
        Xm_La = np.einsum('ii,Lia->La', DELTA2, X_Lia) / 2
        return Xm_La[..., np.newaxis, :]

    dXm_Lia = Property

    def _get_dXm_Lia(self):
        X_Lia = self.X_Ia[self.I_Li]
        return self.Xm_Lia - X_Lia

    def get_vis3d(self):
        X_Lia = self.X_Ia[self.I_Li]
        X_aiL = np.einsum('Lia->aiL', X_Lia)
        X_Lai = np.einsum('Lia->Lai', X_Lia)
        Um_trans_Lia = U_Ia[I_Li]
        Phi_Lia = Phi_Ia[I_Li]
        Um_rot_Lia = np.einsum('abc,...b,...c->...a',
                               EPS, Phi_Lia, self.dXm_Lia)
        Um_Lia = Um_trans_Lia + Um_rot_Lia
        #---
        Um_trans_aiL = np.einsum('Lia->aiL', Um_trans_Lia)
        XU_aiL = X_aiL + Um_trans_aiL
        XUm_aiL = np.einsum('Lia->aiL', (self.Xm_Lia + Um_Lia))
        XUIm_aiL = np.concatenate(
            [
                np.einsum('iaL->aiL',
                          np.array([XU_aiL[:, 0, ...], XUm_aiL[:, 0, ...]])),
                np.einsum('iaL->aiL',
                          np.array([XU_aiL[:, 1, ...], XUm_aiL[:, 1, ...]]))
            ], axis=-1
        )
        XU_Lai = np.einsum('aiL->Lai', XU_aiL)
        XUm_Lai = np.einsum('aiL->Lai', XUm_aiL)
        XUIm_Lai = np.einsum('aiL->Lai', XUIm_aiL)

        return X_Lai, XU_Lai, XUm_Lai, XUIm_Lai


def test01_spring():
    X_Ia = np.array(
        [
            [0, 0, 0],
            [0, 1, 0],
        ], dtype=np.float64
    )

    I_Li = np.array(
        [
            [0, 1],
        ], dtype=np.int_
    )

    U_Ia = np.array(
        [
            [0, 0, 0],
            [0, 0, 0],
        ], dtype=np.float64
    )

    Phi_Ia = np.array(
        [
            [0, 0, 0.0 * np.pi],
            [0, 0, 0.0 * np.pi],
        ], dtype=np.float64
    )

    fixed_dofs = [0, 1, 2, 3, 4, 5,
                  8, 9, 10, 11]
    control_dofs = []
    return X_Ia, I_Li, U_Ia, Phi_Ia, fixed_dofs, control_dofs


def test02_quad():
    X_Ia = np.array(
        [
            [0, 0, 0],
            [0, 2, 0],
            [-1, 1, 0],
            [1, 1, 0]
        ], dtype=np.float64
    )

    I_Li = np.array(
        [
            [0, 2],
            [0, 3],
            [1, 2],
            [1, 3]
        ], dtype=np.int_
    )

    U_Ia = np.array(
        [
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0]
        ], dtype=np.float64
    )

    Phi_Ia = np.array(
        [
            [0, 0, 0.1 * np.pi],
            [0, 0, -0.1 * np.pi],
            [0, 0, 0.1 * np.pi],
            [0.3 * np.pi, 0, 0.1 * np.pi]
        ], dtype=np.float64
    )

    fixed_dofs = []
    control_dofs = []

    return X_Ia, I_Li, U_Ia, Phi_Ia, fixed_dofs, control_dofs


def test03_tetrahedron():
    X_Ia = np.array(
        [
            [0, 0, 0],
            [1, 0, 0],
            [0, 1, 0],
            [0.3, 0.3, 1.0]
        ], dtype=np.float64
    )

    I_Li = np.array(
        [
            [0, 1],
            [0, 2],
            [0, 3],
            [1, 2],
            [1, 3],
            [2, 3]
        ], dtype=np.int_
    )

    U_Ia = np.array(
        [
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0]
        ], dtype=np.float64
    )

    Phi_Ia = np.array(
        [
            [0, 0, 0.1 * np.pi],
            [0, 0, -0.1 * np.pi],
            [0, 0, 0.1 * np.pi],
            [0.3 * np.pi, 0, 0.1 * np.pi]
        ], dtype=np.float64
    )

    fixed_dofs = [0, 1, 2, 3, 4, 5,
                  7, 8, 9, 10, 11,
                  12, 14, 15, 16, 17,
                  21, 23]

    control_dofs = [22]

    return X_Ia, I_Li, U_Ia, Phi_Ia, fixed_dofs, control_dofs


def test04_pyramide():
    X_Ia = np.array(
        [
            [0, 0, 0],
            [1, 0, 0],
            [0, 1, 0],
            [1, 1, 0],
            [0.5, 0.5, 1.0],
        ], dtype=np.float64
    )

    I_Li = np.array(
        [
            [0, 1],
            [0, 2],
            [0, 3],
            [1, 2],
            [1, 3],
            [2, 3],
            [0, 4],
            [1, 4],
            [2, 4],
            [3, 4]
        ], dtype=np.int_
    )

    U_Ia = np.array(
        [
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [0.02, 0.03, 0]
        ], dtype=np.float64
    )

    Phi_Ia = np.array(
        [
            [0, 0, 0.01 * np.pi],
            [0, 0, -0.01 * np.pi],
            [0, 0, 0.01 * np.pi],
            [0, 0, -0.01 * np.pi],
            [0.03 * np.pi, 0, 0.01 * np.pi]
        ], dtype=np.float64
    )

    fixed_dofs = []
    control_dofs = []
    return X_Ia, I_Li, U_Ia, Phi_Ia, fixed_dofs, control_dofs


def plot3d(X_Ia, I_Li):
    X_Lia = X_Ia[I_Li]
    X_Lai = np.einsum('Lia->Lai', X_Lia)
    plt.figure()
    ax = plt.axes(projection="3d")
    for X_ai in X_Lai:
        x_line, y_line, z_line = X_ai
        ax.plot3D(x_line, y_line, z_line, 'gray')
    x_points, y_points, z_points = np.einsum('Ia->aI', X_Ia)
    ax.scatter3D(x_points, y_points, z_points, color='red')
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')


def plot_u3d(X_Lai, XU_Lai, XUm_Lai, XUIm_Lai):
    # def plot_u3d(X_Lia, XU_Lia):

    plt.figure()
    ax = plt.axes(projection="3d")
    for X_ai, XU_ai, XUm_ai in zip(X_Lai, XU_Lai, XUm_Lai):
        #    for X_ai, XU_ai in zip(X_Lai, XU_Lai):
        x_line, y_line, z_line = X_ai
        ax.plot3D(x_line, y_line, z_line, 'gray')
        x_line, y_line, z_line = XU_ai
        ax.plot3D(x_line, y_line, z_line, 'green')
        x_line, y_line, z_line = XUm_ai
        ax.plot3D(x_line, y_line, z_line, 'red')
    for XUIm_ai in XUIm_Lai:
        x_line, y_line, z_line = XUIm_ai
        ax.plot3D(x_line, y_line, z_line, 'blue')
#     x_points, y_points, z_points = np.einsum('Ia->aI', XU_Ia)
#     ax.scatter3D(x_points, y_points, z_points, color='red')
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')


if __name__ == '__main__':
    X_Ia, I_Li, U_Ia, Phi_Ia, fixed_dofs, control_dofs = test03_tetrahedron()
    lt = LatticeTessellation(X_Ia=X_Ia, I_Li=I_Li)
    xdomain = XDomainLattice(mesh=lt)
    U_pLia = np.array([U_Ia[I_Li], Phi_Ia[I_Li]])
    U = np.einsum('pLia->Lipa', U_pLia).flatten()
    eps_La = xdomain.map_U_to_field(U)

    from ibvpy.tmodel.mats3D_ifc import MATS3DIfcElastic
    mats = MATS3DIfcElastic()
    sig_La, D_Lab = mats.get_corr_pred(eps_La, 1.0)
    L_O, F_Lo = xdomain.map_field_to_F(sig_La)
    F_int = np.bincount(L_O, weights=F_Lo)

    k_mtx_array = xdomain.map_field_to_K(D_Lab)
    K = SysMtxAssembly()
    K.sys_mtx_arrays.append(k_mtx_array)
    K_dense = DenseMtx(assemb=K)
    print(K_dense)

    for dof in fixed_dofs:
        K.register_constraint(a=dof, u_a=0)

    K.register_constraint(a=22,  u_a=0.3)  # clamped end
    F_ext = np.zeros_like(F_int)
    F_ext[20] = -100000
    K.apply_constraints(F_ext)
    # print(F_ext)
    #K_dense = DenseMtx(assemb=K)
    print('Rank 2', np.linalg.matrix_rank(K_dense.mtx))
    U, pd = K.solve()
    U_Ia = U[xdomain.o_Ipa][..., 0, :]
    Phi_Ia = U[xdomain.o_Ipa][..., 1, :]

    eps_La = xdomain.map_U_to_field(U)
    sig_La, D_Lab = mats.get_corr_pred(eps_La, 1.0)
    L_O, F_Lo = xdomain.map_field_to_F(sig_La)
    F_int = np.bincount(L_O, weights=F_Lo)
    #print('F_int 2\n', F_int)

    #Fu = np.einsum('ij,j->i', K_dense.mtx, U)
    # print(Fu)

    X_Lai, XU_Lai, XUm_Lai, XUIm_Lai = xdomain.get_vis3d()
    plot_u3d(X_Lai, XU_Lai, XUm_Lai, XUIm_Lai)
    plt.show()



#-------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Sep 4, 2009 by: rch

from ibvpy.bcond.bcond_mngr import BCondMngr
from ibvpy.sim.tstep_bc import TStepBC
from ibvpy.tmodel.mats_eval import IMATSEval
from traits.api import \
    Instance, \
    Dict, WeakRef, List, \
    DelegatesTo, Bool

import numpy as np


# from pyglet.media.drivers.alsa.asound import u_int16_t
class MATSXDExplore(TStepBC):

    '''
    Base class for MATSExplorer dimensional to manage dimensionally 
    dependent presentation of the material models.
    Simulate the loading histories of a material point in 1D space.
        '''

    explorer = WeakRef

    node_name = 'Stress space'

    tree_node_list = List([])

    def _tree_node_list_default(self):
        return [
            self.mats_eval
        ]

    def _update_node_list(self):
        print('updating Stress space', self.mats_eval)
        self.tree_node_list = [
            self.mats_eval
        ]

    state_array_shapes = DelegatesTo('mats_eval')

    algorithmic_stiffness = Bool(
        True, ALG=True, auto_set=False, enter_set=True)

    # Boundary condition manager
    #
    bcond_mngr = Instance(BCondMngr)

    def _bcond_mngr_default(self):
        return BCondMngr()

    def get_corr_pred(self, U, dU, t_n, t_n1, update_state,
                      **state_vars):
        print('U', U)
        eps_Emab = self.mats_eval.map_eps_eng_to_mtx(U)[np.newaxis, ...]
        deps_Emab = self.mats_eval.map_eps_eng_to_mtx(dU)[np.newaxis, ...]
        D_Emabef, sig_Emab = self.mats_eval.get_corr_pred(
            eps_Emab, deps_Emab, t_n, t_n1,
            update_state, self.algorithmic_stiffness,
            **state_vars
        )
        K = self.mats_eval.map_tns4_to_tns2(D_Emabef[0])
        F_int = self.mats_eval.map_sig_mtx_to_eng(sig_Emab[0])
        print('K', K)
        print('F', F_int)
        return K, F_int



'''
Created on Dec 17, 2018

@author: rch
'''

from traits.api import \
    provides, List, Dict, WeakRef, Property, cached_property
from ibvpy.view.plot2d import Vis2D
import numpy as np
import copy
from .i_hist import IHist


@provides(IHist)
class Hist(Vis2D):
    '''Object storing and managing the history of calculation.
    '''

    tstep_source = WeakRef

    timesteps = List()

    U_list = List()
    F_list = List()

    vis_record = Dict

    record_dict = Property(
        Dict, depends_on='tstep_source.record, tstep_source.record_items'
    )

    @cached_property
    def _get_record_dict(self):
        ts = self.tstep_source
        for vis in self.vis_record.values():
            vis.tstep = ts
        return {key: vis for key, vis in self.vis_record.items()}

    def __getitem__(self, key):
        return self.record_dict[key]

    state_vars = List()

    def init_state(self):
        self.timesteps = []
        self.U_list = []
        self.F_list = []
        self.state_vars = []
        # self.Eps_t = {}
        for vis in self.record_dict.values():
            vis.setup()

    def record_timestep(self, t, U, F,
                        state_vars=None):
        '''Add the time step and record the 
        corresponding state of the model.
        '''
        self.timesteps.append(t)
        self.U_list.append(np.copy(U))
        self.F_list.append(np.copy(F))
        self.state_vars.append(copy.deepcopy(state_vars))
        for vis in self.record_dict.values():
            vis.update()

    # def insert_Eps(self, Eps):
    #     for Eps_d in Eps:
    #
    #     self.Eps_t
    #     keys = self.Eps_t[0,0].keys()
    #     Eps_t = self.Eps_t[idx,0]
    #     if reduce_dim:
    #         eps_tEms = eps_tEms[0,...]
    #         Eps_Dt = Eps_t[0]
    #     else:
    #         Eps_Dt = {
    #             key: np.array([Eps[key] for i, Eps in enumerate(Eps_t)], dtype=np.float64)
    #             for key in keys
    #         }

    U_t = Property(depends_on='timesteps_items')

    @cached_property
    def _get_U_t(self):
        return np.array(self.U_list, dtype=np.float64)

    F_t = Property(depends_on='timesteps_items')

    @cached_property
    def _get_F_t(self):
        return np.array(self.F_list, dtype=np.float64)

    # todo: capture the mapping between multidomain state representation
    #       and a time slice during evaluation
    Eps_t = Property(depends_on='timesteps_items')
    '''Array of state variable evolution'''
    @cached_property
    def _get_Eps_t(self):
        return np.array(self.state_vars, dtype=np.object)

    t = Property(depends_on='timesteps_items')

    @cached_property
    def _get_t(self):
        return np.array(self.timesteps, dtype=np.float64)

    n_t = Property(depends_on='timesteps_items')

    @cached_property
    def _get_n_t(self):
        return len(self.timesteps)

    def get_time_idx_arr(self, vot):
        '''Get the index corresponding to visual time
        '''
        t = self.t
        if len(t) == 0:
            return 0
        idx = np.array(np.arange(len(t)), dtype=np.float64) + 0.5
        t_idx = np.interp(vot, t, idx)
        return np.array(t_idx, np.int_)

    def get_time(self, vot):
        return self.t[self.get_time_idx(vot)]

    def get_time_idx(self, vot):
        return int(self.get_time_idx_arr(vot))



from traits.api import \
     Array, Bool, Callable, Enum, Float, HasTraits, Interface, \
     Instance, Int, Trait, Str, Enum, Callable, List, Any, \
     on_trait_change, Tuple, WeakRef, Delegate, Property, cached_property, \
     This, self, TraitError, Dict

from .i_fe_parent_domain import IFEParentDomain
from .fe_subdomain import FESubDomain
from .fe_domain import FEDomain

class FERefinementLevel( FESubDomain ):

    # specialized label
    _tree_label = Str( 'refinement level' )

    def _set_domain( self, value ):
        'reset the domain of this domain'
        if self.parent != None:
            raise TraitError('child FESubDomain cannot be added to FEDomain')
        super( FERefinementLevel, self )._set_domain( value )
    def _get_domain( self ):
        if self.parent != None:
            return self.parent.domain
        return super( FERefinementLevel, self )._get_domain()

    def validate( self ):
        if self.parent != None:
            raise ValueError('only parentless subdomains can be inserted into domain')

    # children domains: list of the instances of the same class
    children = List( This )

    # parent domain
    _parent = This( domain_changed = True )
    parent = Property( This )
    def _set_parent( self, value ):
        'reset the parent of this domain'
        if self._parent:
            # check to see that the changed parent 
            # is within the same domain
            if value.domain != self._parent.domain:
                raise NotImplementedError('Parent change across domains not implemented')
            # unregister in the old parent
            self._parent.children.remove( self )
        else:
            # append the current subdomain at the end of the subdomain
            # series within the domain
            #pass
            value.domain._append_in_series( self )
        # set the new parent
        self._parent = value
        # register the subdomain in the new parent
        self._parent.children.append( self )
    def _get_parent( self ):
        return self._parent

    #---------------------------------------------------------------------------------------
    # Implement the child interface
    #---------------------------------------------------------------------------------------
    # Element refinement representation
    #
    refinement_dict = Dict( changed_structure = True )

    def refine_elem( self, parent_ix, *refinement_args ):
        '''For the specified parent position let the new element decompose.
        '''
        if parent_ix in self.refinement_dict:
            raise ValueError('element %s already refined' % repr(parent_ix))

        # the element is deactivated in the parent domain
        self.refinement_dict[ parent_ix ] = refinement_args
        self.parent.deactivate( parent_ix )




from ibvpy.tmodel.mats2D.mats2D_elastic.mats2D_elastic import MATS2DElastic
from ibvpy.api import FEDomain, FEGrid, FERefinementGrid, TStepper as TS
from ibvpy.fets.fets2D.fets2D4q import FETS2D4Q

if __name__ == '__main__':

    fets_eval_4u = FETS2D4Q(mats_eval = MATS2DElastic())
    
    fe_domain = FEDomain()

    fe_rgrid1 = FERefinementGrid( name = 'fe_rgrid1', fets_eval = fets_eval_4u, domain = fe_domain )

    fe_grid1 = FEGrid( name = 'fe_grid1', coord_max = (2.,6.,0.), 
                               shape   = (1,3),
                               fets_eval = fets_eval_4u,
                               level = fe_rgrid1 )    

    fe_grid2 = FEGrid( name = 'fe_grid2', coord_min = (2.,  6, 0.),
                      coord_max = (10, 15, 0.), 
                               shape   = (3,2),
                               fets_eval = fets_eval_4u,
                               level = fe_rgrid1 )    
    
    print(fe_grid2[ 1, 0 ].elems)
    
    fe_grid2.deactivate( ( 1, 0 ) )
    print('activation map')
    print(fe_grid2.activation_map)

    ts = TS( sdomain = fe_domain )
    
    from ibvpy.plugins.ibvpy_app import IBVPyApp
    ibvpy_app = IBVPyApp( ibv_resource = ts )
    ibvpy_app.main()    


#-------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Aug 7, 2009 by: rchx

from traits.api import TraitType, HasTraits, TraitError

class EitherType(TraitType):

    def __init__(self, names=None, klasses=None, **metadata):
        # validate that these are trait types
        self._klasses = klasses
        self._names = names
        super(EitherType, self).__init__(**metadata)

    def validate(self, object, name, value):
        ''' Set the trait value '''
        # first check if the value is a class
        if isinstance(value, type):
            klass = value
            if not klass in self._klasses:
                raise TraitError('type %s not in the type scope' % klass)
            # check if the last instance of the klass has been
            # registered earlier in the trait history
            new_value = klass()
        else:
            # the value must be one of those in _klasses
            if isinstance(value, tuple(self._klasses)):
                new_value = value
            else:
                raise TraitError('value of type %s out of the scope: %s' %
                                 (value.__class__, self._klasses))
        return new_value

    def get_default_value(self):
        '''Take the first class to construct the value'''
        klass = self._klasses[0]
        value = klass()
        return (0, value)

    def create_editor(self):
        pass